import itertools
from math import sqrt

import torch
import matplotlib.pyplot as plt
import numpy as np

from ftools.ai.calc.f_calc_adv import f_cre_grid_cells


class FAnchors_v1():

    def __init__(self, anc_scale, strides=(8, 16, 32)) -> None:
        super().__init__()
        self.anc_scale = anc_scale
        self.strides = strides
        self.ancs_ltrb_input = None
        self.nums_dim_t_list = None  # 保存每个特图的 h*w 维度
        self.strides_match_input = None  # 用于各层特图和input的转换 ts
        self.size_hw_input = None  # 用于动态判断

    def get_anc_ltrb_input(self, sizes_hw_t, device, size_hw_input, is_update=False):
        if self.size_hw_input != size_hw_input:
            # 尺寸不一样动态
            is_update = True

        if self.ancs_ltrb_input is None or is_update:
            self.ancs_ltrb_input = []
            self.nums_dim_t_list = []  # sizes_hw_t h*w
            self.strides_match_input = []  # 用于各层特图和input的转换
            for i in range(len(self.strides)):
                '''
                获取input的网格点坐标
                anc 的大小是特图5 -> input
                统一到 input 算出anc的ltrbn
                采用ATSS匹配
                '''
                # 保存每个特图的 h*w 维度
                _dim = int(sizes_hw_t[i][0] * sizes_hw_t[i][1])
                self.strides_match_input.extend([self.strides[i]] * _dim)

                self.nums_dim_t_list.append(_dim)
                # base 中心
                x = 0
                y = x
                # base wh
                scale = self.strides[i] * self.anc_scale
                w = 0.5 * (scale - 1)
                h = w
                base_anchor = torch.tensor([x - w, y - h, x + w, y + h], device=device) \
                    .round().view(1, -1).repeat(_dim, 1)

                # torch.Size([2704, 2])
                grid_xy_input = f_cre_grid_cells(sizes_hw_t[i], stride=self.strides[i],
                                                 is_center=True, device=device)
                # self.grids_xy_input.append(grid_xy_input)

                self.ancs_ltrb_input.append(base_anchor + grid_xy_input.repeat(1, 2))

                # 1600,4   400,4     100,4  -> torch.Size([2100, 4])
            # self.ancs_ltrb_input = torch.cat(self.ancs_ltrb_input, 0).clamp(min=0, max=size_hw_input[0])
            self.ancs_ltrb_input = torch.cat(self.ancs_ltrb_input, 0)
            # 这里要确保中点为整数中间
            self.ancs_ltrb_input = self.ancs_ltrb_input \
                                   - torch.tensor([0, 0, 1, 1], device=device, dtype=torch.float).unsqueeze(0)
            self.strides_match_input = torch.tensor(self.strides_match_input, device=device)
        return self.ancs_ltrb_input


# ---------------------- 新代码 -------------------------
class FAnchors:
    def __init__(self, img_in_size, anc_scale, feature_size_steps=None, feature_sizes=None, anchors_clip=True,
                 is_xymid=False, is_real_size=False, device=None):
        '''
        用于动态anc
        :param img_in_size:
        :param anc_scale:
            ANCHORS_SIZE = [
                [[0.13, 0.10666667], [0.06, 0.15733333], [0.036, 0.06006006], ],
                [[0.196, 0.51466667], [0.29, 0.28], [0.12, 0.28], ],
                [[0.81786211, 0.872], [0.374, 0.72266667], [0.612, 0.452], ],
            ]
        :param feature_size_steps: 可以输入下采样 算尺寸  也可以直接输入尺寸
        :param anchors_clip:
        :param is_xymid:
        :param is_real_size:
        '''
        self.anchors_clip = anchors_clip  # 是否剔除超边界---超参
        self.ancs_scale = anc_scale
        self.is_real_size = is_real_size  # 以 anchors_size 的真实尺寸输出 非归一化
        self.is_xymid = is_xymid  # 是否将anc中心点移到格子中间
        # 特征层对应的步距   [8, 16, 32] 原图size/特图size = images/feature_maps = steps
        self.feature_size_steps = feature_size_steps  # 这个定义特图下采样比例

        self.img_in_size = img_in_size
        # 根据预处理后的尺寸及步距 计算每一个特图的尺寸
        from math import ceil
        # 根据预处理图片及下采倍数，计算特图尺寸 <class 'list'>: [[52, 52], [26, 26], [13, 13]] (w,h)
        if feature_sizes is not None:
            self.feature_sizes = np.array(feature_sizes)[:, None].repeat(2, axis=1)
        else:
            self.feature_sizes = [[ceil(self.img_in_size[0] / step), ceil(self.img_in_size[1] / step)]
                                  for step in self.feature_size_steps]
        self.nums_level = []  # 每层个数
        self.nums_anc = []  # 每层anc数
        self.device = device
        self.ancs_xywh = self.cre_anchors()

        self.match_grids = None
        self.ancs_xywh_t = None

    def cre_anchors(self):
        '''
        形成的顺序是 每一个特图->每个格子(行优先) 建议从小到大 需与模型匹配 ->每一个尺寸
        :return:
            返回特图 h*w,4 anchors 是每个特图的长宽个 4维整框
            这里是 x,y,w,h 调整系数
        '''
        device = self.device
        ret_ancs = torch.empty((0, 4), device=device)
        for i, (row, col) in enumerate(self.feature_sizes):
            # 求xy 中心点坐标 行列与xy是反的
            rowcol_index = f_cre_grid_cells([row, col], stride=1, is_swap=True, num_repeat=1)
            rowcol_index = rowcol_index.to(device)
            # 这个特图的数量
            num_anc_one = rowcol_index.shape[0]  # 2704
            # 这个特图的尺寸 [[0.025, 0.025], [0.05, 0.05]]
            ancs_wh = self.ancs_scale[i]
            # 单体复制 每一个格子有
            num_anc = len(ancs_wh)
            self.nums_anc.append(num_anc)
            self.nums_level.append(num_anc_one * num_anc)  # self.nums_feature 是全新
            rowcol_index = rowcol_index.repeat_interleave(num_anc, dim=0)  # 单体复制
            rowcol = torch.tensor((row, col), device=device)
            # xy 与 rowcol相反
            ancs_yx = torch.true_divide(rowcol_index, rowcol)  # 特图归一化
            if self.is_xymid:  # 移动中心点到格子中间
                midyx = torch.true_divide(1, rowcol) / 2
                ancs_yx = ancs_yx + midyx
            ancs_wh = torch.tensor(ancs_wh, device=device).repeat(num_anc_one, 1)
            ancs_xy = ancs_yx[:, [1, 0]]
            ans_xywh = torch.cat([ancs_xy, ancs_wh], dim=-1)
            ret_ancs = torch.cat([ret_ancs, ans_xywh], dim=0)
        if self.anchors_clip:  # 对于xywh 来说这个参数 是没有用的
            ret_ancs = xywh2ltrb(ret_ancs)
            ret_ancs.clamp_(min=0, max=1)  # 去除超边际的
            ret_ancs = ltrb2xywh(ret_ancs)
        if self.is_real_size:
            ret_ancs = ret_ancs * torch.tensor(self.img_in_size)[None].repeat(1, 2)
        # __d = 1
        if self.device is not None:
            ret_ancs = ret_ancs.to(self.device)
        return ret_ancs

    def match_anc_grids(self):
        # anc 归一化 -> anc特图
        if self.match_grids is not None:
            return self.match_grids, self.ancs_xywh_t

        nums_ceng_np = np.array(self.nums_level, dtype=np.int32)
        nums_anc_np = np.array(self.nums_anc, dtype=np.float32)
        grids_np = np.sqrt((nums_ceng_np / nums_anc_np))  # [ 52 26 13 7 4]
        match_grids_ts = torch.tensor(np.repeat(grids_np, nums_ceng_np, axis=-1), device=self.ancs_xywh.device,
                                      dtype=torch.float)
        match_grids_ts = match_grids_ts.view(-1, 1)  # torch.Size([32526, 1]) -> [32526, 4]
        self.match_grids = match_grids_ts
        self.ancs_xywh_t = self.ancs_xywh * match_grids_ts
        return self.match_grids, self.ancs_xywh_t


class DefaultBoxes:
    def __init__(self, img_size, feature_size, steps, wh, aspect_ratios, scale_xy=0.1, scale_wh=0.2,
                 device=torch.device('cpu')):
        '''

        :param img_size:  输入图片的尺寸
        :param feature_size: 特图尺寸 [38, 19, 10, 5, 3, 1]
        :param steps: 输入图片的下采样比例 向上取整得 特图尺寸
        :param wh: 每个框的真实尺寸
        :param aspect_ratios: anc比例 2 表示 2:1   [2,3]表示2:1 3:1 两个  1比1 是默认就有
        :param scale_xy:
        :param scale_wh:
        '''
        self.img_size = img_size  # 输入网络的图像大小 300
        # [38, 19, 10, 5, 3, 1]
        self.feature_size = feature_size  # 每个预测层的feature map尺寸

        self.scale_xy_ = scale_xy
        self.scale_wh_ = scale_wh

        # According to https://github.com/weiliu89/caffe
        # Calculation method slightly different from paper
        # [8, 16, 32, 64, 100, 300]
        self.steps = steps  # 每个特征层上的一个cell在原图上的跨度

        # [21, 45, 99, 153, 207, 261, 315]
        self.wh = wh  # 每个特征层上预测的default box的scale

        fk = img_size / np.array(steps)  # 计算每层特征层的  正常应该=feature_size
        # [[2], [2, 3], [2, 3], [2, 3], [2], [2]]
        self.aspect_ratios = aspect_ratios  # 每个预测特征层上预测的default box的ratios

        self.default_boxes = []
        # size of feature and number of feature
        # 遍历每层特征层，计算default box
        for idx, sfeat in enumerate(self.feature_size):
            sk1 = wh[idx] / img_size  # scale转为相对值[0-1]
            sk2 = wh[idx + 1] / img_size  # scale转为相对值[0-1]
            sk3 = sqrt(sk1 * sk2)
            # 先添加两个1:1比例的default box宽和高
            all_sizes = [(sk1, sk1), (sk3, sk3)]  # 1:1的默认添加

            # 再将剩下不同比例的default box宽和高添加到all_sizes中
            for alpha in aspect_ratios[idx]:
                w, h = sk1 * sqrt(alpha), sk1 / sqrt(alpha)
                all_sizes.append((w, h))
                all_sizes.append((h, w))

            # 计算当前特征层对应原图上的所有default box
            for w, h in all_sizes:
                for i, j in itertools.product(range(sfeat), repeat=2):  # i -> 行（y）， j -> 列（x）
                    # 计算每个default box的中心坐标（范围是在0-1之间）
                    cx, cy = (j + 0.5) / fk[idx], (i + 0.5) / fk[idx]
                    self.default_boxes.append((cx, cy, w, h))

        # 将default_boxes转为tensor格式
        self.ancs_xywh = torch.as_tensor(self.default_boxes, dtype=torch.float32, device=device)  # 这里不转类型会报错
        self.ancs_xywh.clamp_(min=0, max=1)  # 将坐标（x, y, w, h）都限制在0-1之间

        # For IoU calculation
        # ltrb is left top coordinate and right bottom coordinate
        # 将(x, y, w, h)转换成(xmin, ymin, xmax, ymax)，方便后续计算IoU(匹配正负样本时)
        self.dboxes_ltrb = self.ancs_xywh.clone()
        self.dboxes_ltrb[:, 0] = self.ancs_xywh[:, 0] - 0.5 * self.ancs_xywh[:, 2]  # xmin
        self.dboxes_ltrb[:, 1] = self.ancs_xywh[:, 1] - 0.5 * self.ancs_xywh[:, 3]  # ymin
        self.dboxes_ltrb[:, 2] = self.ancs_xywh[:, 0] + 0.5 * self.ancs_xywh[:, 2]  # xmax
        self.dboxes_ltrb[:, 3] = self.ancs_xywh[:, 1] + 0.5 * self.ancs_xywh[:, 3]  # ymax

    @property
    def scale_xy(self):
        return self.scale_xy_

    @property
    def scale_wh(self):
        return self.scale_wh_

    def __call__(self, order='ltrb'):
        # 根据需求返回对应格式的default box
        if order == 'ltrb':
            return self.dboxes_ltrb

        if order == 'xywh':
            return self.ancs_xywh


def cre_ssd_ancs(device=torch.device('cpu')):
    img_size = 300  # 输入网络的图像大小
    feature_size = [38, 19, 10, 5, 3, 1]  # 每个预测层的feature map尺寸
    steps = [8, 16, 32, 64, 100, 300]  # 每个特征层上的一个cell在原图上的跨度
    # use the scales here: https://github.com/amdegroot/ssd.pytorch/blob/master/data/config.py
    scales = [21, 45, 99, 153, 207, 261, 315]  # 每个特征层上预测的default box的scale
    aspect_ratios = [[2], [2, 3], [2, 3], [2, 3], [2], [2]]  # 每个预测特征层上预测的default box的ratios

    # fig_size = 4
    # feat_size = [2]
    # steps = [1]
    #
    # # 尺寸为2的 有2比1 3比1种
    # scales = [2, 1]  # 长宽比
    # # 2比1 3比1
    # aspect_ratios = [[2, 3]]

    dboxes = DefaultBoxes(img_size, feature_size, steps, scales, aspect_ratios, device=device)
    return dboxes


def tfanc():
    global size, len
    # __t001()
    # __t_anc4found()
    size = [640, 640]
    # anchors_size = [
    #     [[10, 13], [16, 30], [33, 23]],  # 大特图小目标 52, 52
    #     [[30, 61], [62, 45], [59, 119]],  # 26, 26
    #     [[116, 90], [156, 198], [373, 326]],  # 小特图大目标 13x13
    # ]
    feature_map_steps = [8, 16, 32]
    # anchors = Anchors(size, anchors_size, feature_map_steps,
    #                   is_xymid=False, is_real_size=True, anchors_clip=False).get_anchors()  # torch.Size([10647, 4]
    # anc_scale = [
    #     [[0.13, 0.10666667], [0.06, 0.15733333], [0.036, 0.06006006], ],
    #     [[0.196, 0.51466667], [0.29, 0.28], [0.12, 0.28], ],
    #     [[0.81786211, 0.872], [0.374, 0.72266667], [0.612, 0.452], ],
    # ]
    anc_scale = [
        [[0.025, 0.025], [0.05, 0.05]],
        [[0.1, 0.1], [0.2, 0.2], ],
        [[0.4, 0.4], [0.8, 0.8], ],
    ]
    anchors_xywh = FAnchors(size, anc_scale, feature_map_steps,
                            anchors_clip=False, is_xymid=True, is_real_size=False).cre_anchors()
    # print(anchors)
    # anchors = FAnchors(size, anc_scale, feature_map_steps,
    #                    anchors_clip=True, is_xymid=False, is_real_size=True).cre_anchors()  # torch.Size([10647, 4])
    # [[52, 52], [26, 26], [13, 13]]
    index_start = 0
    # index_start = len(anchors) - 460
    len = 6
    anchors_xywh = anchors_xywh[index_start:index_start + len]  # 这里选出 anchors
    # --------------anchors 转换画图--------------
    # __anchors = anchors.clone()
    # __anchors = xywh2ltrb(__anchors)
    # __anchors[:, ::2] = __anchors[:, ::2] * size[0]
    # __anchors[:, 1::2] = __anchors[:, 1::2] * size[1]
    # xywh --> ltwh 为了plt.Rectangle
    # __anchors[:, :2] = __anchors[:, :2] - __anchors[:, 2:] / 2.
    # img_pil_new = Image.new('RGB', list(np.array(size) * 2), (128, 128, 128))
    # img_pil_new = Image.new('RGB', size, (128, 128, 128))
    # # show_bbox4pil(img_pil_new, xywh2ltrb(__anchors))
    # show_bbox4pil(img_pil_new, xywh2ltrb(anchors_xywh))
    # 构造点
    w = [i for i in range(size[0])]
    h = [i for i in range(size[1])]
    ww, hh = np.meshgrid(w, h)
    p = 100  # 图片显示padding
    # ----------图形配置---------
    fig = plt.figure()
    ax = fig.add_subplot(111)
    plt.ylim(-p, size[0] + p)
    plt.xlim(-p, size[1] + p)
    ax.invert_yaxis()  # y轴反向
    plt.scatter(ww.reshape(-1), hh.reshape(-1))
    for a in anchors_xywh:  # 画矩形框
        # xy,w,h
        rect = plt.Rectangle((a[0], a[1]), a[2], a[3],
                             color="r",
                             fill=False)
        ax.add_patch(rect)
    plt.show()
    pass


if __name__ == '__main__':
    # tfanc()

    anc_obj = cre_ssd_ancs()
    anchors_xywh = anc_obj.ancs_xywh
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.invert_yaxis()
    p = 1
    plt.ylim(-p, p)
    plt.xlim(-p, p)
    for a in xywh2ltrb(anchors_xywh):  # 画矩形框
        # ltrb
        rect = plt.Rectangle((a[0], a[1]), a[2], a[3],
                             color="r",
                             fill=False)
        ax.add_patch(rect)
    plt.show()
    print()



import torch
from torch import nn

from ftools.ai.fmodels.backbone.net_blocks import BaseConv, Bottleneck, DWConv, SPPBottleneck


class CSPLayer(nn.Module):
    """C3 in yolov5, CSP Bottleneck with 3 convolutions"""

    def __init__(
            self, in_channels, out_channels, n=1,
            shortcut=True, expansion=0.5, depthwise=False, act="silu"
    ):
        """
        Args:
            in_channels (int): input channels.
            out_channels (int): output channels.
            n (int): number of Bottlenecks. Default value: 1.
        """
        # ch_in, ch_out, number, shortcut, groups, expansion
        super().__init__()
        hidden_channels = int(out_channels * expansion)  # hidden channels
        self.conv1 = BaseConv(in_channels, hidden_channels, 1, stride=1, act=act)
        self.conv2 = BaseConv(in_channels, hidden_channels, 1, stride=1, act=act)
        self.conv3 = BaseConv(2 * hidden_channels, out_channels, 1, stride=1, act=act)
        module_list = [
            Bottleneck(hidden_channels, hidden_channels, shortcut, 1.0, depthwise, act=act)
            for _ in range(n)
        ]
        self.m = nn.Sequential(*module_list)

    def forward(self, x):
        x_1 = self.conv1(x)
        x_2 = self.conv2(x)
        x_1 = self.m(x_1)
        x = torch.cat((x_1, x_2), dim=1)
        return self.conv3(x)


class Focus(nn.Module):
    """Focus width and height information into channel space."""

    def __init__(self, in_channels, out_channels, ksize=1, stride=1, act="silu"):
        super().__init__()
        self.conv = BaseConv(in_channels * 4, out_channels, ksize, stride, act=act)

    def forward(self, x):
        # shape of x (b,c,w,h) -> y(b,4c,w/2,h/2)
        # patch_top_left = x[..., ::2, ::2]
        # patch_top_right = x[..., ::2, 1::2]
        # patch_bot_left = x[..., 1::2, ::2]
        # patch_bot_right = x[..., 1::2, 1::2]

        j02_2 = torch.arange(0, x.shape[2], 2)
        j12_2 = torch.arange(1, x.shape[2], 2)
        j02_3 = torch.arange(0, x.shape[3], 2)
        j12_3 = torch.arange(1, x.shape[3], 2)
        patch_top_left = x[..., j02_2, :][..., j02_3]
        patch_top_right = x[..., j02_2, :][..., j12_3]
        patch_bot_left = x[..., j12_2, :][..., j02_3]
        patch_bot_right = x[..., j12_2, :][..., j12_3]

        x = torch.cat(
            (patch_top_left, patch_bot_left, patch_top_right, patch_bot_right,), dim=1,
        )
        return self.conv(x)


class CSPDarknet(nn.Module):

    def __init__(self, dep_mul, wid_mul,
                 out_features=("dark3", "dark4", "dark5"),
                 depthwise=False, act="silu",
                 ):
        super().__init__()
        assert out_features, "please provide output features of Darknet"
        self.out_features = out_features
        Conv = DWConv if depthwise else BaseConv

        base_channels = int(wid_mul * 64)  # 64
        base_depth = max(round(dep_mul * 3), 1)  # 3

        # stem
        self.stem = Focus(3, base_channels, ksize=3, act=act)

        # dark2
        self.dark2 = nn.Sequential(
            Conv(base_channels, base_channels * 2, 3, 2, act=act),
            CSPLayer(
                base_channels * 2, base_channels * 2,
                n=base_depth, depthwise=depthwise, act=act
            ),
        )

        # dark3
        self.dark3 = nn.Sequential(
            Conv(base_channels * 2, base_channels * 4, 3, 2, act=act),
            CSPLayer(
                base_channels * 4, base_channels * 4,
                n=base_depth * 3, depthwise=depthwise, act=act,
            ),
        )

        # dark4
        self.dark4 = nn.Sequential(
            Conv(base_channels * 4, base_channels * 8, 3, 2, act=act),
            CSPLayer(
                base_channels * 8, base_channels * 8,
                n=base_depth * 3, depthwise=depthwise, act=act,
            ),
        )

        # dark5
        self.dark5 = nn.Sequential(
            Conv(base_channels * 8, base_channels * 16, 3, 2, act=act),
            SPPBottleneck(base_channels * 16, base_channels * 16, activation=act),
            CSPLayer(
                base_channels * 16, base_channels * 16, n=base_depth,
                shortcut=False, depthwise=depthwise, act=act,
            ),
        )

    def forward(self, x):
        outputs = {}
        x = self.stem(x)
        outputs["stem"] = x
        x = self.dark2(x)
        outputs["dark2"] = x
        x = self.dark3(x)
        outputs["dark3"] = x
        x = self.dark4(x)
        outputs["dark4"] = x
        x = self.dark5(x)
        outputs["dark5"] = x

        outs_dict = {k: v for k, v in outputs.items() if k in self.out_features}
        return list(outs_dict.values())
        # return outputs["dark3"], outputs["dark4"], outputs["dark5"]


class YOLOXFPN(nn.Module):
    def __init__(self, in_channels, depth=1.0, width=1.0, ):
        super().__init__()
        act = "silu"
        self.upsample = nn.Upsample(scale_factor=2, mode="nearest")
        self.lateral_conv0 = BaseConv(
            int(in_channels[2] * width), int(in_channels[1] * width), 1, 1, act=act
        )

    def forward(self, input):
        pass


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
import os
import numpy as np
import cv2
import shutil
import sys

# sys.path.append(os.path.abspath(os.path.join(os.getcwd(), "..")))
# from pfld.utils import calculate_pitch_yaw_roll
debug = False

import math


def calculate_pitch_yaw_roll(landmarks_2D, cam_w=256, cam_h=256,
                             radians=False):
    """ Return the the pitch  yaw and roll angles associated with the input image.
    @param radians When True it returns the angle in radians, otherwise in degrees.
    """

    assert landmarks_2D is not None, 'landmarks_2D is None'

    # Estimated camera matrix values.
    c_x = cam_w / 2
    c_y = cam_h / 2
    f_x = c_x / np.tan(60 / 2 * np.pi / 180)
    f_y = f_x
    camera_matrix = np.float32([[f_x, 0.0, c_x],
                                [0.0, f_y, c_y],
                                [0.0, 0.0, 1.0]])
    camera_distortion = np.float32([0.0, 0.0, 0.0, 0.0, 0.0])

    # dlib (68 landmark) trached points
    # TRACKED_POINTS = [17, 21, 22, 26, 36, 39, 42, 45, 31, 35, 48, 54, 57, 8]
    # wflw(98 landmark) trached points
    # TRACKED_POINTS = [33, 38, 50, 46, 60, 64, 68, 72, 55, 59, 76, 82, 85, 16]
    # X-Y-Z with X pointing forward and Y on the left and Z up.
    # The X-Y-Z coordinates used are like the standard coordinates of ROS (robotic operative system)
    # OpenCV uses the reference usually used in computer vision:
    # X points to the right, Y down, Z to the front
    landmarks_3D = np.float32([
        [6.825897, 6.760612, 4.402142],  # LEFT_EYEBROW_LEFT,
        [1.330353, 7.122144, 6.903745],  # LEFT_EYEBROW_RIGHT,
        [-1.330353, 7.122144, 6.903745],  # RIGHT_EYEBROW_LEFT,
        [-6.825897, 6.760612, 4.402142],  # RIGHT_EYEBROW_RIGHT,
        [5.311432, 5.485328, 3.987654],  # LEFT_EYE_LEFT,
        [1.789930, 5.393625, 4.413414],  # LEFT_EYE_RIGHT,
        [-1.789930, 5.393625, 4.413414],  # RIGHT_EYE_LEFT,
        [-5.311432, 5.485328, 3.987654],  # RIGHT_EYE_RIGHT,
        [-2.005628, 1.409845, 6.165652],  # NOSE_LEFT,
        [-2.005628, 1.409845, 6.165652],  # NOSE_RIGHT,
        [2.774015, -2.080775, 5.048531],  # MOUTH_LEFT,
        [-2.774015, -2.080775, 5.048531],  # MOUTH_RIGHT,
        [0.000000, -3.116408, 6.097667],  # LOWER_LIP,
        [0.000000, -7.415691, 4.070434],  # CHIN
    ])
    landmarks_2D = np.asarray(landmarks_2D, dtype=np.float32).reshape(-1, 2)

    # Applying the PnP solver to find the 3D pose of the head from the 2D position of the landmarks.
    # retval - bool
    # rvec - Output rotation vector that, together with tvec, brings points from the world coordinate system to the camera coordinate system.
    # tvec - Output translation vector. It is the position of the world origin (SELLION) in camera co-ords
    _, rvec, tvec = cv2.solvePnP(landmarks_3D, landmarks_2D,
                                 camera_matrix, camera_distortion)
    # Get as input the rotational vector, Return a rotational matrix

    # const double PI = 3.141592653;
    # double thetaz = atan2(r21, r11) / PI * 180;
    # double thetay = atan2(-1 * r31, sqrt(r32*r32 + r33*r33)) / PI * 180;
    # double thetax = atan2(r32, r33) / PI * 180;

    rmat, _ = cv2.Rodrigues(rvec)
    pose_mat = cv2.hconcat((rmat, tvec))
    _, _, _, _, _, _, euler_angles = cv2.decomposeProjectionMatrix(pose_mat)
    return map(lambda k: k[0], euler_angles)  # euler_angles contain (pitch, yaw, roll)


def rotate(angle, center, landmark):
    rad = angle * np.pi / 180.0
    alpha = np.cos(rad)
    beta = np.sin(rad)
    M = np.zeros((2, 3), dtype=np.float32)
    M[0, 0] = alpha
    M[0, 1] = beta
    M[0, 2] = (1 - alpha) * center[0] - beta * center[1]
    M[1, 0] = -beta
    M[1, 1] = alpha
    M[1, 2] = beta * center[0] + (1 - alpha) * center[1]

    landmark_ = np.asarray([(M[0, 0] * x + M[0, 1] * y + M[0, 2],
                             M[1, 0] * x + M[1, 1] * y + M[1, 2]) for (x, y) in landmark])
    return M, landmark_


class ImageDate:
    def __init__(self, line, path_img, image_size=112):
        self.image_size = image_size
        line = line.strip().split()
        # 0-195: landmark 坐标点  196-199: bbox 坐标点;
        # 200: 姿态(pose)         0->正常姿态(normal pose)          1->大的姿态(large pose)
        # 201: 表情(expression)   0->正常表情(normal expression)    1->夸张的表情(exaggerate expression)
        # 202: 照度(illumination) 0->正常照明(normal illumination)  1->极端照明(extreme illumination)
        # 203: 化妆(make-up)      0->无化妆(no make-up)             1->化妆(make-up)
        # 204: 遮挡(occlusion)    0->无遮挡(no occlusion)           1->遮挡(occlusion)
        # 205: 模糊(blur)         0->清晰(clear)                    1->模糊(blur)
        # 206: 图片名称
        assert (len(line) == 207)
        self.list = line
        self.landmark = np.asarray(list(map(float, line[:196])), dtype=np.float32).reshape(-1, 2)
        self.box = np.asarray(list(map(int, line[196:200])), dtype=np.int32)
        flag = list(map(int, line[200:206]))
        flag = list(map(bool, flag))
        self.pose = flag[0]
        self.expression = flag[1]
        self.illumination = flag[2]
        self.make_up = flag[3]
        self.occlusion = flag[4]
        self.blur = flag[5]
        self.path = os.path.join(path_img, line[206])
        self.img = None

        self.imgs = []
        self.landmarks = []
        self.boxes = []

    def load_data(self, is_train, repeat, mirror=None):
        if (mirror is not None):
            with open(mirror, 'r') as f:
                lines = f.readlines()
                assert len(lines) == 1
                mirror_idx = lines[0].strip().split(',')
                mirror_idx = list(map(int, mirror_idx))
        xy = np.min(self.landmark, axis=0).astype(np.int32)
        zz = np.max(self.landmark, axis=0).astype(np.int32)
        wh = zz - xy + 1

        center = (xy + wh / 2).astype(np.int32)
        img = cv2.imread(self.path)
        boxsize = int(np.max(wh) * 1.2)
        xy = center - boxsize // 2
        x1, y1 = xy
        x2, y2 = xy + boxsize
        height, width, _ = img.shape
        dx = max(0, -x1)
        dy = max(0, -y1)
        x1 = max(0, x1)
        y1 = max(0, y1)

        edx = max(0, x2 - width)
        edy = max(0, y2 - height)
        x2 = min(width, x2)
        y2 = min(height, y2)

        imgT = img[y1:y2, x1:x2]
        if (dx > 0 or dy > 0 or edx > 0 or edy > 0):
            imgT = cv2.copyMakeBorder(imgT, dy, edy, dx, edx, cv2.BORDER_CONSTANT, 0)
        if imgT.shape[0] == 0 or imgT.shape[1] == 0:
            imgTT = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
            for x, y in (self.landmark + 0.5).astype(np.int32):
                cv2.circle(imgTT, (x, y), 1, (0, 0, 255))
            cv2.imshow('0', imgTT)
            if cv2.waitKey(0) == 27:
                exit()
        imgT = cv2.resize(imgT, (self.image_size, self.image_size))
        landmark = (self.landmark - xy) / boxsize
        assert (landmark >= 0).all(), str(landmark) + str([dx, dy])
        assert (landmark <= 1).all(), str(landmark) + str([dx, dy])
        self.imgs.append(imgT)
        self.landmarks.append(landmark)

        if is_train:
            while len(self.imgs) < repeat:
                angle = np.random.randint(-30, 30)
                cx, cy = center
                cx = cx + int(np.random.randint(-boxsize * 0.1, boxsize * 0.1))
                cy = cy + int(np.random.randint(-boxsize * 0.1, boxsize * 0.1))
                M, landmark = rotate(angle, (cx, cy), self.landmark)

                imgT = cv2.warpAffine(img, M, (int(img.shape[1] * 1.1), int(img.shape[0] * 1.1)))

                wh = np.ptp(landmark, axis=0).astype(np.int32) + 1
                size = np.random.randint(int(np.min(wh)), np.ceil(np.max(wh) * 1.25))
                xy = np.asarray((cx - size // 2, cy - size // 2), dtype=np.int32)
                landmark = (landmark - xy) / size
                if (landmark < 0).any() or (landmark > 1).any():
                    continue

                x1, y1 = xy
                x2, y2 = xy + size
                height, width, _ = imgT.shape
                dx = max(0, -x1)
                dy = max(0, -y1)
                x1 = max(0, x1)
                y1 = max(0, y1)

                edx = max(0, x2 - width)
                edy = max(0, y2 - height)
                x2 = min(width, x2)
                y2 = min(height, y2)

                imgT = imgT[y1:y2, x1:x2]
                if (dx > 0 or dy > 0 or edx > 0 or edy > 0):
                    imgT = cv2.copyMakeBorder(imgT, dy, edy, dx, edx, cv2.BORDER_CONSTANT, 0)

                imgT = cv2.resize(imgT, (self.image_size, self.image_size))

                if mirror is not None and np.random.choice((True, False)):
                    landmark[:, 0] = 1 - landmark[:, 0]
                    landmark = landmark[mirror_idx]
                    imgT = cv2.flip(imgT, 1)
                self.imgs.append(imgT)
                self.landmarks.append(landmark)

    def save_data(self, path, prefix):
        attributes = [self.pose, self.expression, self.illumination, self.make_up, self.occlusion, self.blur]
        attributes = np.asarray(attributes, dtype=np.int32)
        attributes_str = ' '.join(list(map(str, attributes)))
        labels = []
        TRACKED_POINTS = [33, 38, 50, 46, 60, 64, 68, 72, 55, 59, 76, 82, 85, 16]
        for i, (img, lanmark) in enumerate(zip(self.imgs, self.landmarks)):
            assert lanmark.shape == (98, 2)
            save_path = os.path.join(path, prefix + '_' + str(i) + '.png')
            assert not os.path.exists(save_path), save_path
            cv2.imwrite(save_path, img)

            euler_angles_landmark = []
            for index in TRACKED_POINTS:
                euler_angles_landmark.append(lanmark[index])
            euler_angles_landmark = np.asarray(euler_angles_landmark).reshape((-1, 28))
            pitch, yaw, roll = calculate_pitch_yaw_roll(euler_angles_landmark[0])
            euler_angles = np.asarray((pitch, yaw, roll), dtype=np.float32)
            euler_angles_str = ' '.join(list(map(str, euler_angles)))

            landmark_str = ' '.join(list(map(str, lanmark.reshape(-1).tolist())))

            label = '{} {} {} {}\n'.format(save_path, landmark_str, attributes_str, euler_angles_str)

            labels.append(label)
        return labels


def get_dataset_list(imgDir, outDir, landmarkDir, is_train):
    with open(landmarkDir, 'r') as f:
        lines = f.readlines()
        labels = []
        save_img = os.path.join(outDir, 'imgs')
        if not os.path.exists(save_img):
            os.mkdir(save_img)

        if debug:
            lines = lines[:100]
        for i, line in enumerate(lines):
            Img = ImageDate(line, imgDir)
            img_name = Img.path
            Img.load_data(is_train, 10, Mirror_file)
            _, filename = os.path.split(img_name)
            filename, _ = os.path.splitext(filename)
            label_txt = Img.save_data(save_img, str(i) + '_' + filename)
            labels.append(label_txt)
            if ((i + 1) % 100) == 0:
                print('file: {}/{}'.format(i + 1, len(lines)))

    with open(os.path.join(outDir, 'list.txt'), 'w') as f:
        for label in labels:
            f.writelines(label)


if __name__ == '__main__':
    root_dir = os.path.dirname(os.path.realpath(__file__))
    imageDirs = 'WFLW/WFLW_images'
    Mirror_file = 'WFLW/WFLW_annotations/Mirror98.txt'

    landmarkDirs = ['WFLW/WFLW_annotations/list_98pt_rect_attr_train_test/list_98pt_rect_attr_test.txt',
                    'WFLW/WFLW_annotations/list_98pt_rect_attr_train_test/list_98pt_rect_attr_train.txt']

    outDirs = ['test_data', 'train_data']
    for landmarkDir, outDir in zip(landmarkDirs, outDirs):
        outDir = os.path.join(root_dir, outDir)
        print(outDir)
        if os.path.exists(outDir):
            shutil.rmtree(outDir)
        os.mkdir(outDir)
        if 'list_98pt_rect_attr_test.txt' in landmarkDir:
            is_train = False
        else:
            is_train = True
        imgs = get_dataset_list(imageDirs, outDir, landmarkDir, is_train)
    print('end')



import onnxruntime
import torch
from torchvision import models

from ftools.ai.fmodels.f_layer_get import ModelOut4Resnet18
from yolov1.net.net_yolov1 import YOLOv1_NetOD


def ffun_onnx(pretrained_dict_y):
    dd = {}
    for k, v in pretrained_dict_y.items():
        dd[k.replace('net.', '')] = v
    return dd


if __name__ == '__main__':
    import os

    from ftools.ai.CONFIG_BASE import FCFG_BASE
    from ftools.ai.fits.fweight import load_weight

    cfg = FCFG_BASE()
    cfg.device = torch.device('cpu')
    cfg.SIZE_WH_INPUT_TRAIN  = (320, 320)

    cfg.NUM_CLASSES = 3
    cfg.STRIDE = 32

    backbone = models.resnet18(pretrained=True)
    backbone = ModelOut4Resnet18(backbone)
    model = YOLOv1_NetOD(backbone, cfg)

    _file_read_weight = None  # 无yolo1文件
    # path_root_weight = r'M:\AI\weights\feadre\zz'
    # _file_read_weight = os.path.join(path_root_weight, 'nanodet_shuffle2_type3_38_12.03_p75.7_r63.4' + '.pth')

    start_epoch = load_weight(file_weight=_file_read_weight, model=model,
                              optimizer=None, lr_scheduler=None,
                              device=cfg.device, ffun=ffun_onnx)

    batch = 1
    num_end = 10
    data_ts = torch.rand((batch, 3, *cfg.SIZE_WH_INPUT_TRAIN ), dtype=torch.float)
    reses = model(data_ts)

    for res in reses:
        print(res.reshape(batch, -1)[:, :num_end])

    # 验证
    file_onnx = 'yolov1.onnx'
    input_names = ['input']
    output_names = ['out_cls', 'out_conf_box']

    torch.onnx.export(model, data_ts, file_onnx, verbose=False,
                      input_names=input_names,
                      output_names=output_names,
                      dynamic_axes=None,
                      opset_version=11,
                      )

    onnx_session = onnxruntime.InferenceSession(file_onnx)
    res_onnx = onnx_session.run(
        output_names=output_names,
        input_feed={input_names[0]: data_ts.numpy()})

    for res in res_onnx:
        print(res.reshape(batch, -1)[:, :num_end])
    print('debug')



import os

import torch

from ftools.GLOBAL_LOG import flog

'''
filter(lambda p: p.requires_grad, model.parameters()) 过滤优化器的迭代器
'''


def save_weight(path_save, model, name, loss=None, optimizer=None,
                lr_scheduler=None, epoch=1, maps_val=None,
                ema_model=None, ):
    if path_save and os.path.exists(path_save):

        if ema_model is not None:
            model = ema_model.ema

        sava_dict = {
            'model': model.state_dict(),
            'optimizer': optimizer.state_dict() if optimizer else None,
            'lr_scheduler': lr_scheduler.state_dict() if lr_scheduler else None,
            'epoch': epoch}

        if loss is not None:
            l = round(loss, 2)
        else:
            l = ''

        if maps_val is not None:
            file_weight = os.path.join(path_save, (name + '_{}_{}_{}_{}.pth')
                                       .format(epoch,
                                               l,
                                               'p' + str(round(maps_val[0] * 100, 1)),
                                               'r' + str(round(maps_val[1] * 100, 1)),
                                               ))
        else:
            file_weight = os.path.join(path_save, (name + '_{}_{}.pth').format(epoch, l))
        torch.save(sava_dict, file_weight)
        flog.info('保存成功 %s', file_weight)


def load_weight(file_weight, model, optimizer=None, lr_scheduler=None,
                device=torch.device('cpu'), is_mgpu=False, ffun=None):
    start_epoch = 1
    if file_weight and os.path.exists(file_weight):
        checkpoint = torch.load(file_weight, map_location=device)

        '''对多gpu的k进行修复'''
        if 'model' in checkpoint:
            pretrained_dict_y = checkpoint['model']
        else:
            pretrained_dict_y = checkpoint

        '''重组权重回调'''
        if ffun is not None:
            pretrained_dict = ffun(pretrained_dict_y)
        else:
            pretrained_dict = pretrained_dict_y

        dd = {}
        # # 多GPU处理
        ss = 'module.'
        for k, v in pretrained_dict.items():
            if is_mgpu:
                if ss not in k:
                    dd[ss + k] = v
                else:
                    dd = pretrained_dict_y
                    break
                    # dd[k] = v
            else:
                dd[k.replace(ss, '')] = v

        keys_missing, keys_unexpected = model.load_state_dict(dd, strict=False)
        if len(keys_missing) > 0 or len(keys_unexpected):
            flog.error('missing_keys %s', keys_missing)  # 这个是 model 的属性
            flog.error('unexpected_keys %s', keys_unexpected)  # 这个是 pth 的属性
        if optimizer and 'optimizer' in checkpoint and checkpoint['optimizer'] is not None:
            optimizer.load_state_dict(checkpoint['optimizer'])
        if lr_scheduler and 'lr_scheduler' in checkpoint and checkpoint['lr_scheduler'] is not None:
            lr_scheduler.load_state_dict(checkpoint['lr_scheduler'])

        if 'epoch' in checkpoint:
            start_epoch = checkpoint['epoch'] + 1


        flog.warning('已加载 feadre 权重文件为 %s', file_weight)
    else:
        flog.error('权重文件加载失败 %s', file_weight)
    return start_epoch



import cv2
import torch
import numpy as np
import torchvision
from PIL import Image


def resize_np_keep(img_np, new_size_hw, mode='lt', is_fill=True, fill_color=(128, 128, 128)):
    '''
    
    :param img_np: 一张图片 np图片 h,w,c
    :param new_size_hw: [h,w]
    :param mode: center lt
    :param is_fill: 是否用 fill_color 进行填充
    :param fill_color:
    :return:
        img, old_size list(h,w) 最后进行反向
        (top, bottom, left, right) # 填充宽度
    '''
    assert isinstance(img_np, np.ndarray), 'type(img_np = ' + str(type(img_np))
    old_size_wh = img_np.shape[0:2]  # hw
    # flog.debug('原尺寸 %s', img.shape)
    # 选择比例小的
    ratio = min(float(new_size_hw[i]) / (old_size_wh[i]) for i in range(len(old_size_wh)))
    # flog.debug('选择的缩放比例 %s', ratio)

    # 未填充时的尺寸需要保存下来
    h_new, w_new = tuple([int(i * ratio) for i in old_size_wh])
    # flog.debug('新尺寸 %s %s', h_new, w_new)
    interpolation = cv2.INTER_LINEAR if ratio > 1.0 else cv2.INTER_AREA
    img_np = cv2.resize(img_np, (w_new, h_new), interpolation=interpolation)  # 大坑 W H

    # cv2.imshow('bb', img)
    # cv2.waitKey()
    left, top, right, bottom = None, None, None, None
    if is_fill:

        # 图片填充
        _pad_w = new_size_hw[1] - w_new
        _pad_h = new_size_hw[0] - h_new

        top, bottom, left, right = 0, 0, 0, 0
        if mode == 'center':
            top, bottom = _pad_h // 2, _pad_h - (_pad_h // 2)
            left, right = _pad_w // 2, _pad_w - (_pad_w // 2)
        elif mode == 'lt':
            right = _pad_w
            bottom = _pad_h

        img_np = cv2.copyMakeBorder(img_np, top, bottom, left, right, cv2.BORDER_CONSTANT, None, fill_color)
    return img_np, ratio, old_size_wh[::-1], (left, top, right, bottom)


def resize_img_tensor(image, new_size_wh):
    '''
    将图片缩放到指定的大小范围内，并处理对应缩放bboxes信息
    :param image: 一张已toTensor图片 c,h,w torch.Size([3, 335, 500])
    :param new_size_wh: [w,h]
    :param target: 一个target对象
    :return: tensor c,h,w torch.Size([3, new_size])
    '''
    # ------确定一个缩放因子 以最大尺寸为优先, 缩放图片到指定的宽高--------
    # image shape is [channel, height, width]
    hw = [new_size_wh[1], new_size_wh[0]]
    old_size_hw = image.shape[-2:]  # (800 600) 这个取出来不是tensor
    ratio = min(np.array(hw) / old_size_hw)

    if ratio > 1.0:
        image = torch.nn.functional.interpolate(
            image[None], size=hw,
            mode='bilinear', align_corners=False)
    else:
        image = torch.nn.functional.interpolate(
            image[None], size=hw,
            mode='nearest')

    return image.squeeze(0), old_size_hw[::-1]


def resize_pil_keep(img_pil, new_size, is_fill=True):
    '''
    is_fill==False 自动判断长宽均满足 剪切填充 满足new_size的一边 确保图形的完整
    :param img_pil:
    :param new_size:
    :param is_fill: True 充满图片  False 时并不填充
    :return:
    '''
    old_size = img_pil.size

    if is_fill:
        ratio = max(float(new_size[i]) / (old_size[i]) for i in range(len(old_size)))
        nw, nh = tuple([int(i * ratio) for i in old_size])
        img_pil = img_pil.resize((nw, nh), Image.ANTIALIAS)  # Image.BICUBIC
        img_pil = img_pil.crop((0, 0, *new_size))
    else:
        # 选择比例小的
        ratio = min(float(new_size[i]) / (old_size[i]) for i in range(len(old_size)))
        # flog.debug('选择的缩放比例 %s', ratio)
        # 未填充时的尺寸需要保存下来
        nw, nh = tuple([int(i * ratio) for i in old_size])
        # flog.debug('新尺寸 %s %s', nw, nh)
        img_pil = img_pil.resize((nw, nh), Image.ANTIALIAS)
        # img_pil = img_pil.crop((0, 0, nw, nh))
    return img_pil, ratio, (nw, nh)


def letterbox(img_np, new_shape=(416, 416), color=(128, 128, 128), auto=True, scaleFill=False, scaleup=True):
    # Resize image to a 32-pixel-multiple rectangle https://github.com/ultralytics/yolov3/issues/232
    shape = img_np.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better test mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, 32), np.mod(dh, 32)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = new_shape
        ratio = new_shape[0] / shape[1], new_shape[1] / shape[0]  # width, height ratios

    dw /= 2  # divide padding into 2 sides
    dh /= 2

    if shape[::-1] != new_unpad:  # resize
        img_np = cv2.resize(img_np, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    img_np = cv2.copyMakeBorder(img_np, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    return img_np, ratio, (dw, dh)


if __name__ == '__main__':
    file_pic = r'D:\tb\tb\ai_code\DL\_test_pic\2007_006046.jpg'
    file_pic = r'/_test_pic/test_99.jpg'

    # img = cv2.imread(file_img)
    from torchvision.transforms import functional as F

    # 直接拉伸 pil打开为RGB
    img_pil = Image.open(file_pic).convert('RGB')
    # img_pil.show()
    # img_pil, _, _ = resize_img_pil_keep(img_pil, (416, 416), is_fill=True)
    # img_pil, _, _ = resize_img_pil_keep(img_pil, (416, 600), is_fill=False)
    # img_pil = img_pil.crop((0, 0, 300, 200))

    # interpolation = cv2.INTER_LINEAR if ratio > 1.0 else cv2.INTER_AREA
    # img_pil = img_pil.resize((300, 600), Image.ANTIALIAS)
    # img_pil.show()

    ''' h,w,c(brg) '''
    # img_np = img_np[:, :, ::-1].transpose(2, 0, 1) #  BGR to RGB
    img_np = cv2.cvtColor(np.array(img_pil), cv2.COLOR_RGB2BGR)
    print(img_np.shape)
    # img_np, ratio, old_size_wh, (left, top, right, bottom) = resize_np_keep(img_np, (600, 400), mode='center')
    # print(ratio, old_size_wh, (left, top, right, bottom))
    img_np, ratio, old_size, _ = resize_np_keep(img_np, (416, 416), is_fill=True, )
    # print(ratio, old_size, (left, top, right, bottom))
    # img_np = cv2.imread(file_pic)  # 读取原始图像
    # img_np, ratio, (dw, dh) = letterbox(img_np, scaleup=False)
    # cv2.imshow("img", img_np)  # 显示只支持BGR
    # cv2.waitKey(0)

    img_pil = Image.fromarray(img_np, mode="RGB")
    img_pil.show()

    # img_tensor = F.to_tensor(img_pil)  # Image w,h -> c,h,w +归一化
    # img_tensor, old_size_wh = resize_img_tensor(img_tensor, (600, 200))
    # print(old_size_wh)
    # img_pil = F.to_pil_image(img_tensor).convert('RGB')  # 这个会还原

    # img_pil, ratio, (nw, nh) = resize_pil_keep(img_pil, (600, 200), is_fill=False)
    # print(ratio, (nw, nh))
    # img_pil.show()



import os

from ftools.ai.datas.coco.coco_api import f_look_coco_type, f_coco_verify_info
from ftools.ai.datas.data_factory import fcre_dataload
from ftools.ai.gpu.webapi.translate import translate_baidu

KEYPOINTS_Mirror = {  # 水平翻转 这个要和数据遍历对起
    'face98': {
        0: 32, 1: 31, 2: 30, 3: 29, 4: 28, 5: 27, 6: 26, 7: 25, 8: 24, 9: 23, 10: 22, 11: 21, 12: 20,
        13: 19,
        14: 18,
        15: 17,
        16: 16, 17: 15, 18: 14, 19: 13, 20: 12, 21: 11, 22: 10, 23: 9, 24: 8, 25: 7, 26: 6, 27: 5,
        28: 4, 29: 3,
        30: 2,
        31: 1, 32: 0,
        33: 46, 34: 45, 35: 44, 36: 43, 37: 42, 38: 50, 39: 49, 40: 48, 41: 47,
        46: 33, 45: 34, 44: 35, 43: 36, 42: 37, 50: 38, 49: 39, 48: 40, 47: 41,
        60: 72, 61: 71, 62: 70, 63: 69, 64: 68, 65: 75, 66: 74, 67: 73,
        72: 60, 71: 61, 70: 62, 69: 63, 68: 64, 75: 65, 74: 66, 73: 67,
        96: 97, 97: 96,
        51: 51, 52: 52, 53: 53, 54: 54,
        55: 59, 56: 58, 57: 57, 58: 56, 59: 55,
        76: 82, 77: 81, 78: 80, 79: 79, 80: 78, 81: 77, 82: 76,
        87: 83, 86: 84, 85: 85, 84: 86, 83: 87,
        88: 92, 89: 91, 90: 90, 91: 89, 92: 88,
        95: 93, 94: 94, 93: 95
    },
    'face5': [[0, 5], [3, 4]],  # widerface
    'lsp': [[0, 5], [1, 4], [2, 3], [6, 11], [7, 10], [8, 9]],  # 14个关键点
}

KEYPOINTS_SEQ = {
    'LSP': [[13, 12], [12, 9], [12, 8], [9, 10], [8, 7], [10, 11],
            [7, 6], [12, 3], [12, 2], [2, 1], [1, 0], [3, 4], [4, 5]],
}


class DataInfo:

    def __init__(self, dataset_name, path_root,
                 mode_train, batch_train, path_img_train, file_json_train,
                 mode_test, batch_test, path_img_test, file_json_test,
                 kps_seq=None,
                 ) -> None:
        '''

        '''
        super().__init__()
        self.dataset_name = dataset_name  # 数据集的通用名称不能区分是train 还是test
        self.path_root = path_root
        self.kps_seq = kps_seq

        self.mode_train = mode_train
        self.batch_train = batch_train
        self.path_img_train = path_img_train
        self.file_json_train = file_json_train

        self.mode_test = mode_train if mode_test is None else mode_test
        self.batch_test = batch_train if batch_test is None else batch_test
        self.path_img_test = path_img_test
        self.file_json_test = file_json_test


def get_dataloader(cfg, data_info, transform_train, transform_test, kps_seq, ):
    dataloader_train, dataloader_test = None, None
    if cfg.IS_TRAIN and data_info.file_json_train is not None:
        dataloader_train = fcre_dataload(is_multi_scale=cfg.IS_MULTI_SCALE,
                                         num_workers=cfg.NUM_WORKERS,
                                         mode=data_info.mode_train,
                                         batch=data_info.batch_train,
                                         file_json=data_info.file_json_train,
                                         path_img=data_info.path_img_train,
                                         transform=transform_train,
                                         is_jump=cfg.IS_JUMP,
                                         name=data_info.dataset_name + '_train',
                                         size_hw_img=cfg.SIZE_WH_INPUT_TRAIN,
                                         is_mosaic=cfg.IS_MOSAIC,
                                         is_mixup=cfg.IS_MIXUP,
                                         is_mosaic_filter=cfg.IS_MOSAIC_FILTER,
                                         thr_wh=cfg.THR_WH,
                                         thr_area=cfg.THR_AREA,
                                         kps_seq=kps_seq,
                                         is_debug=cfg.IS_DEBUG,
                                         )

        f_look_coco_type(dataloader_train.dataset.coco_obj,
                         ids_cats_ustom=None,
                         name=dataloader_train.dataset.name)

    if (cfg.IS_VAL or cfg.IS_TEST) and data_info.file_json_test is not None:
        dataloader_test = fcre_dataload(is_multi_scale=cfg.IS_MULTI_SCALE,
                                        num_workers=cfg.NUM_WORKERS,
                                        mode=data_info.mode_test,
                                        batch=data_info.batch_test,
                                        file_json=data_info.file_json_test,
                                        path_img=data_info.path_img_test,
                                        transform=transform_test,
                                        is_jump=False,
                                        name=data_info.dataset_name + '_test',
                                        kps_seq=kps_seq,
                                        is_debug=cfg.IS_DEBUG,
                                        )
        f_look_coco_type(dataloader_test.dataset.coco_obj,
                         ids_cats_ustom=None,
                         name=dataloader_test.dataset.name)
    return dataloader_train, dataloader_test


def get_data_cocomin(cfg, path_host,
                     mode_train, batch_train, transform_train=None,
                     mode_test=None, batch_test=None, transform_test=None
                     ):
    ''' 这个可用于COCO测试 '''
    cfg.NUMS_VAL_DICT = {55: 3, 66: 1, 77: 1}  # 验证起始和频率
    cfg.NUMS_TEST_DICT = {1: 1, 100: 1, 200: 1}  # 测试的起始和频率
    cfg.NUM_WEIGHT_SAVE_DICT = {1: 1, 35: 1, 50: 1}
    cfg.PRINT_FREQ = 30
    cfg.VAL_FREQ = 20
    cfg.NUM_CLASSES = 80  # 这个根据数据集要改

    dataset_name = 'cocomin'
    path_root = os.path.join(path_host, '/AI/datas/coco2017')
    path_img_train = os.path.join(path_root, 'imgs/train2017_118287')
    file_json_train = os.path.join(path_root, 'annotations/instances_coco_min_train_16000.json')
    path_img_test = os.path.join(path_root, 'imgs/val2017_5000')
    file_json_test = os.path.join(path_root, 'annotations/instances_coco_min_val_3200.json')
    # file_json_test = os.path.join(path_root, 'annotations/instances_val2017_4952.json')
    kps_seq = None

    data_info = DataInfo(dataset_name, path_root,
                         mode_train, batch_train, path_img_train, file_json_train,
                         mode_test, batch_test, path_img_test, file_json_test,
                         kps_seq=kps_seq,
                         )

    dataloader_train, dataloader_test = get_dataloader(
        cfg, data_info,
        transform_train, transform_test,
        kps_seq=kps_seq,
    )

    return data_info, dataloader_train, dataloader_test


def get_data_coco2017(cfg, path_host,
                      mode_train, batch_train, transform_train=None,
                      mode_test=None, batch_test=None, transform_test=None
                      ):
    ''' 这个可用于COCO测试 '''
    cfg.NUMS_VAL_DICT = {55: 3, 66: 1, 77: 1}  # 验证起始和频率
    cfg.NUMS_TEST_DICT = {1: 1, 100: 1, 200: 1}  # 测试的起始和频率
    cfg.NUM_WEIGHT_SAVE_DICT = {1: 1, 35: 1, 50: 1}
    cfg.PRINT_FREQ = 40
    cfg.VAL_FREQ = 40
    cfg.NUM_CLASSES = 80  # 这个根据数据集要改

    dataset_name = 'coco2017'
    path_root = os.path.join(path_host, '/AI/datas/coco2017')
    path_img_train = os.path.join(path_root, 'imgs/train2017_118287')
    file_json_train = os.path.join(path_root, 'annotations/instances_train2017_117266.json')
    path_img_test = os.path.join(path_root, 'imgs/val2017_5000')
    file_json_test = os.path.join(path_root, 'annotations/instances_val2017_4952.json')
    # file_json_test = os.path.join(path_root, 'annotations/instances_val2017_4952.json')
    kps_seq = None

    data_info = DataInfo(dataset_name, path_root,
                         mode_train, batch_train, path_img_train, file_json_train,
                         mode_test, batch_test, path_img_test, file_json_test,
                         kps_seq=kps_seq,
                         )

    dataloader_train, dataloader_test = get_dataloader(
        cfg, data_info,
        transform_train, transform_test,
    )

    return data_info, dataloader_train, dataloader_test


def get_data_type3(cfg, path_host,
                   mode_train, batch_train, transform_train=None,
                   mode_test=None, batch_test=None, transform_test=None
                   ):
    cfg.NUMS_VAL_DICT = {55: 3, 66: 1, 77: 1}  # 验证起始和频率
    cfg.NUMS_TEST_DICT = {20: 1, 100: 1, 200: 1}  # 测试的起始和频率
    cfg.NUM_WEIGHT_SAVE_DICT = {15: 10, 35: 1, 50: 1}  # 25轮记录
    cfg.PRINT_FREQ = 8
    cfg.NUM_CLASSES = 3  # ** 这里要改

    dataset_name = 'type3'  # ** 这里要改
    path_root = os.path.join(path_host, '/AI/datas/VOC2007')
    path_img_train = os.path.join(path_root, 'train/JPEGImages')
    file_json_train = os.path.join(path_root, 'annotations/instances_type3_train_1066.json')
    path_img_test = os.path.join(path_root, 'val/JPEGImages')
    # instances_type3_val_413.json
    file_json_test = os.path.join(path_root, 'annotations/instances_type3_test_637.json')
    kps_seq = None

    # 这里不变
    data_info = DataInfo(dataset_name, path_root,
                         mode_train, batch_train, path_img_train, file_json_train,
                         mode_test, batch_test, path_img_test, file_json_test,
                         kps_seq=kps_seq,
                         )
    kps_seq = None

    dataloader_train, dataloader_test = get_dataloader(
        cfg, data_info,
        transform_train, transform_test,
        kps_seq=kps_seq,
    )

    return data_info, dataloader_train, dataloader_test


def get_data_type4(cfg, path_host,
                   mode_train, batch_train, transform_train=None,
                   mode_test=None, batch_test=None, transform_test=None
                   ):
    cfg.NUMS_VAL_DICT = {55: 3, 66: 1, 77: 1}  # 验证起始和频率
    cfg.NUMS_TEST_DICT = {20: 1, 100: 1, 200: 1}  # 测试的起始和频率
    cfg.NUM_WEIGHT_SAVE_DICT = {15: 10, 35: 1, 50: 1}  # 25轮记录
    cfg.PRINT_FREQ = 8
    cfg.NUM_CLASSES = 4  # ** 这里要改

    dataset_name = 'type4'  # ** 这里要改
    path_root = os.path.join(path_host, '/AI/datas/VOC2007')
    path_img_train = os.path.join(path_root, 'train/JPEGImages')
    file_json_train = os.path.join(path_root, 'annotations/instances_type4_train_994.json')
    path_img_test = os.path.join(path_root, 'val/JPEGImages')
    # instances_type4_test_550.json
    file_json_test = os.path.join(path_root, 'annotations/instances_type4_test_550.json')
    kps_seq = None

    # 这里不变
    data_info = DataInfo(dataset_name, path_root,
                         mode_train, batch_train, path_img_train, file_json_train,
                         mode_test, batch_test, path_img_test, file_json_test,
                         kps_seq=kps_seq,
                         )

    dataloader_train, dataloader_test = get_dataloader(
        cfg, data_info,
        transform_train, transform_test,
        kps_seq=kps_seq,
    )

    return data_info, dataloader_train, dataloader_test


def get_data_face5(cfg, path_host,
                   mode_train, batch_train, transform_train=None,
                   mode_test=None, batch_test=None, transform_test=None
                   ):
    cfg.NUMS_VAL_DICT = {55: 3, 66: 1, 77: 1}  # 验证起始和频率
    cfg.NUMS_TEST_DICT = {1: 1, 100: 1, 200: 1}  # 测试的起始和频率
    cfg.NUM_WEIGHT_SAVE_DICT = {15: 10, 35: 1, 50: 1}  # 25轮记录
    cfg.PRINT_FREQ = 20
    cfg.NUM_CLASSES = 1  # ** 这里要改

    dataset_name = 'face5'  # ** 这里要改
    path_root = os.path.join(path_host, '/AI/datas/kps/face_5')
    path_img_train = os.path.join(path_root, 'images_13466')
    file_json_train = os.path.join(path_root, 'annotations/keypoints_train_10000_10000.json')
    path_img_test = os.path.join(path_root, 'images_13466')
    file_json_test = os.path.join(path_root, 'annotations/keypoints_test_3466_3466.json')

    kps_seq = None
    # 这里不变
    data_info = DataInfo(dataset_name, path_root,
                         mode_train, batch_train, path_img_train, file_json_train,
                         mode_test, batch_test, path_img_test, file_json_test,
                         kps_seq=kps_seq,
                         )

    dataloader_train, dataloader_test = get_dataloader(
        cfg, data_info,
        transform_train, transform_test,
        kps_seq=kps_seq,
    )

    return data_info, dataloader_train, dataloader_test


def get_data_face98(cfg, path_host,
                    mode_train, batch_train, transform_train=None,
                    mode_test=None, batch_test=None, transform_test=None
                    ):
    cfg.NUMS_VAL_DICT = {55: 3, 66: 1, 77: 1}  # 验证起始和频率
    cfg.NUMS_TEST_DICT = {1: 1, 100: 1, 200: 1}  # 测试的起始和频率
    cfg.NUM_WEIGHT_SAVE_DICT = {15: 10, 35: 1, 50: 1}  # 25轮记录
    cfg.PRINT_FREQ = 10
    cfg.NUM_CLASSES = 1  # ** 这里要改

    dataset_name = 'face98'  # ** 这里要改
    path_root = os.path.join(path_host, '/AI/datas/kps/face_98')
    path_img_train = os.path.join(path_root, 'images_train_5316')
    file_json_train = os.path.join(path_root, 'annotations/keypoints_train_7500_5316.json')
    path_img_test = os.path.join(path_root, 'images_test_2118')
    file_json_test = os.path.join(path_root, 'annotations/keypoints_test_2500_2118.json')

    kps_seq = None
    # 这里不变
    data_info = DataInfo(dataset_name, path_root,
                         mode_train, batch_train, path_img_train, file_json_train,
                         mode_test, batch_test, path_img_test, file_json_test,
                         kps_seq=kps_seq,
                         )

    dataloader_train, dataloader_test = get_dataloader(
        cfg, data_info,
        transform_train, transform_test,
        kps_seq=kps_seq,
    )

    return data_info, dataloader_train, dataloader_test


def get_data_voc2007(cfg, path_host,
                     mode_train, batch_train, transform_train=None,
                     mode_test=None, batch_test=None, transform_test=None
                     ):
    cfg.NUMS_VAL_DICT = {55: 3, 66: 1, 77: 1}  # 验证起始和频率
    cfg.NUMS_TEST_DICT = {1: 1, 100: 1, 200: 1}  # 测试的起始和频率
    cfg.NUM_WEIGHT_SAVE_DICT = {15: 10, 35: 1, 50: 1}  # 25轮记录
    cfg.PRINT_FREQ = 20
    cfg.NUM_CLASSES = 20  # ** 这里要改

    dataset_name = 'voc2007'  # ** 这里要改
    path_root = os.path.join(path_host, '/AI/datas/VOC2007')
    path_img_train = os.path.join(path_root, 'train/JPEGImages')
    file_json_train = os.path.join(path_root, 'annotations/instances_train_5011.json')
    path_img_test = os.path.join(path_root, 'val/JPEGImages')
    file_json_test = os.path.join(path_root, 'annotations/instances_val_1980.json')
    kps_seq = None

    # 这里不变
    data_info = DataInfo(dataset_name, path_root,
                         mode_train, batch_train, path_img_train, file_json_train,
                         mode_test, batch_test, path_img_test, file_json_test,
                         kps_seq=kps_seq,
                         )

    dataloader_train, dataloader_test = get_dataloader(
        cfg, data_info,
        transform_train, transform_test,
        kps_seq=kps_seq,
    )

    return data_info, dataloader_train, dataloader_test


def get_data_voc2012(cfg, path_host,
                     mode_train, batch_train, transform_train=None,
                     mode_test=None, batch_test=None, transform_test=None
                     ):
    cfg.NUMS_VAL_DICT = {55: 3, 66: 1, 77: 1}  # 验证起始和频率
    cfg.NUMS_TEST_DICT = {1: 1, 100: 1, 200: 1}  # 测试的起始和频率
    cfg.NUM_WEIGHT_SAVE_DICT = {15: 10, 35: 1, 50: 1}  # 25轮记录
    cfg.PRINT_FREQ = 30
    cfg.NUM_CLASSES = 20  # ** 这里要改

    dataset_name = 'voc2012'  # ** 这里要改
    path_root = os.path.join(path_host, '/AI/datas/VOC2012')
    path_img_train = os.path.join(path_root, 'train/JPEGImages')
    file_json_train = os.path.join(path_root, 'annotations/instances_train_17125.json')
    path_voc2007 = cfg.PATH_HOST + '/AI/datas/VOC2007'
    path_img_test = os.path.join(path_voc2007, 'val/JPEGImages')
    file_json_test = os.path.join(path_voc2007, 'annotations/instances_test_2972.json')

    kps_seq = None
    # 这里不变
    data_info = DataInfo(dataset_name, path_root,
                         mode_train, batch_train, path_img_train, file_json_train,
                         mode_test, batch_test, path_img_test, file_json_test,
                         kps_seq=kps_seq,
                         )

    dataloader_train, dataloader_test = get_dataloader(
        cfg, data_info,
        transform_train, transform_test,
        kps_seq=kps_seq,
    )

    return data_info, dataloader_train, dataloader_test


def get_data_widerface(cfg, path_host,
                       mode_train, batch_train, transform_train=None,
                       mode_test=None, batch_test=None, transform_test=None
                       ):
    cfg.NUMS_VAL_DICT = {55: 3, 66: 1, 77: 1}  # 验证起始和频率
    cfg.NUMS_TEST_DICT = {1: 1, 100: 1, 200: 1}  # 测试的起始和频率
    cfg.NUM_WEIGHT_SAVE_DICT = {15: 10, 35: 1, 50: 1}  # 25轮记录
    cfg.PRINT_FREQ = 30
    cfg.NUM_CLASSES = 1  # ** 这里要改

    dataset_name = 'widerface'  # ** 这里要改
    path_root = os.path.join(path_host, '/AI/datas/widerface')
    path_img_train = os.path.join(path_root, 'images/train2017')
    file_json_train = os.path.join(path_root, 'annotations/person_keypoints_train2017.json')
    path_img_test = os.path.join(path_root, 'images/val2017')
    file_json_test = os.path.join(path_root, 'annotations/instances_val2017.json')

    kps_seq = None
    # 这里不变
    data_info = DataInfo(dataset_name, path_root,
                         mode_train, batch_train, path_img_train, file_json_train,
                         mode_test, batch_test, path_img_test, file_json_test,
                         kps_seq=kps_seq,
                         )

    dataloader_train, dataloader_test = get_dataloader(
        cfg, data_info,
        transform_train, transform_test,
        kps_seq=kps_seq,
    )

    return data_info, dataloader_train, dataloader_test


def get_data_lsp(cfg, path_host,
                 mode_train, batch_train, transform_train=None,
                 mode_test=None, batch_test=None, transform_test=None
                 ):
    cfg.NUMS_VAL_DICT = {55: 3, 66: 1, 77: 1}  # 验证起始和频率
    cfg.NUMS_TEST_DICT = {1: 1, 100: 1, 200: 1}  # 测试的起始和频率
    cfg.NUM_WEIGHT_SAVE_DICT = {15: 10, 35: 1, 50: 1}  # 25轮记录
    cfg.PRINT_FREQ = 30
    cfg.NUM_CLASSES = 1  # ** 这里要改

    dataset_name = 'lsp'  # ** 这里要改
    path_root = os.path.join(path_host, '/AI/datas/kps/Leeds_Sports_Pose')
    path_img_train = os.path.join(path_root, 'data/lsp_dataset/images')
    file_json_train = os.path.join(path_root, 'annotations/keypoints_train_1600_1600.json')
    path_img_test = os.path.join(path_root, 'data/lsp_dataset/images')
    file_json_test = os.path.join(path_root, 'annotations/keypoints_test_400_400.json')

    kps_seq = KEYPOINTS_SEQ['LSP']
    # 这里不变
    data_info = DataInfo(dataset_name, path_root,
                         mode_train, batch_train, path_img_train, file_json_train,
                         mode_test, batch_test, path_img_test, file_json_test,
                         kps_seq=kps_seq,
                         )

    dataloader_train, dataloader_test = get_dataloader(
        cfg, data_info,
        transform_train, transform_test,
        kps_seq=kps_seq,
    )

    return data_info, dataloader_train, dataloader_test


def get_data_raccoon(cfg, path_host,
                     mode_train, batch_train, transform_train=None,
                     mode_test=None, batch_test=None, transform_test=None
                     ):
    cfg.NUMS_VAL_DICT = {55: 3, 66: 1, 77: 1}  # 验证起始和频率
    cfg.NUMS_TEST_DICT = {1: 1, 100: 1, 200: 1}  # 测试的起始和频率
    cfg.NUM_WEIGHT_SAVE_DICT = {15: 10, 35: 1, 50: 1}  # 25轮记录
    cfg.PRINT_FREQ = 10
    cfg.NUM_CLASSES = 1  # ** 这里要改

    dataset_name = 'raccoon'  # ** 这里要改
    path_root = os.path.join(path_host, '/AI/datas/raccoon200')
    path_img_train = os.path.join(path_root, 'VOCdevkit/JPEGImages')
    file_json_train = os.path.join(path_root, 'annotations/instances_train2017.json')
    path_img_test = os.path.join(path_root, 'VOCdevkit/JPEGImages')
    file_json_test = os.path.join(path_root, 'annotations/instances_val2017.json')

    # 这里不变
    data_info = DataInfo(dataset_name, path_root,
                         mode_train, batch_train, path_img_train, file_json_train,
                         mode_test, batch_test, path_img_test, file_json_test,
                         kps_seq=None,
                         )

    dataloader_train, dataloader_test = get_dataloader(
        cfg, data_info,
        transform_train, transform_test,
        kps_seq=None,
    )

    return data_info, dataloader_train, dataloader_test


def cre_id_classes():
    ids_classes_china = {}
    for k, v in ids_classes.items():
        ids_classes_china[k] = translate_baidu(v)
        time.sleep(1)
    # 创建id class文件
    file_json = os.path.join(data_info.path_root, 'ids_classes_%s.json' % data_info.dataset_name)
    with open(file_json, 'w', encoding='utf-8') as f:
        json.dump(ids_classes, f, ensure_ascii=False, )
    file_json = os.path.join(data_info.path_root, 'classes_ids_%s.json' % data_info.dataset_name)
    with open(file_json, 'w', encoding='utf-8') as f:
        json.dump(dataset.classes_ids, f, ensure_ascii=False, )
    # 创建中文的id class文件
    file_json = os.path.join(data_info.path_root, 'ids_classes_%s_china.json' % data_info.dataset_name)
    with open(file_json, 'w', encoding='utf-8') as f:
        json.dump(ids_classes_china, f, ensure_ascii=False, )
    print('创建成功 ', file_json)


if __name__ == '__main__':
    '''
    这里为每一个数据集创建 class id文件
    '''
    import json
    import time


    class CFG:
        pass


    path_host = 'M:'
    cfg = CFG()
    cfg.IS_TRAIN = True
    cfg.IS_VAL = False
    cfg.IS_TEST = True
    cfg.IS_MULTI_SCALE = False
    cfg.SIZE_WH_INPUT_TRAIN = (320, 320)
    cfg.IS_MOSAIC = False
    cfg.IS_MIXUP = False
    cfg.IS_MOSAIC_FILTER = False
    cfg.THR_WH = 5
    cfg.THR_AREA = 25
    cfg.NUM_WORKERS = 0
    mode_train = 'keypoints'  # bbox segm keypoints caption
    # get_data_widerface 数据中自带校验方法
    data_info, dataloader_train, dataloader_test = get_data_lsp(cfg, path_host=path_host,
                                                                mode_train=mode_train,
                                                                batch_train=10)

    dataset = dataloader_train.dataset
    ids_classes = dataset.ids_classes

    for data in dataset:
        data1 = data
        pass

    ''' ------------------------------- 创建文件 ------------------------------- '''
    # 创建3个 id_classes 的 josn
    # cre_id_classes()
    #
    ''' ------------------------------- 简单测试数据集 ------------------------------- '''
    # print('dataloader_test')
    # f_coco_verify_info(dataloader_test.dataset.coco_obj, data_info.path_img_train)
    #
    # print('dataloader_train')
    # f_coco_verify_info(dataloader_train.dataset.coco_obj, data_info.path_img_test)



import random

import onnx
import onnxruntime
import torch
from torch import nn
import numpy as np
from torchvision import models


def fcre_onnx(model, data_ts, file_onnx, verbose=False,
              input_names=('input'),  # 单输入
              output_names=('output'),
              dynamic_axes=None,
              ):
    '''
    # 选择用那种ops.有三种选择：
    # 1. OperatorExportTypes.ONNX（默认），导出为标准ONNX运算符
    # 2. OperatorExportTypes.ONNX_ATEN， 导出为aten运算符
    # 3. OperatorExportTypes.ONNX_ATEN_FALLBACK， 在ONNX中注册的运算符使用ONNX,其余使用aten
    :param model:
    :param data_ts:
    :param file_onnx:
    :param verbose:
    :param input_names:
    :param output_names:
    :param dynamic_axes:
    :return:
    '''
    torch.onnx.export(model,
                      data_ts,
                      file_onnx,
                      export_params=True,  # 是否保存模型的训练好的参数
                      verbose=False,  # 是否输出debug描述
                      input_names=['input1', 'input2'],  # 定义输入结点的名字，有几个输入就定义几个
                      output_names=['output'],  # 定义输出结点的名字
                      opset_version=11,  # onnx opset的库版本
                      do_constant_folding=True,  # whether do constant-folding optimization 该优化会替换全为常数输入的分支为最终结果
                      operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
    torch.onnx.export(model, data_ts, file_onnx, verbose=verbose,
                      input_names=input_names,
                      output_names=output_names,
                      dynamic_axes=dynamic_axes,
                      )


class TModel(nn.Module):

    def __init__(self):
        super().__init__()

    def forward(self, inputs):
        # export 自动 eval
        x, y = list(inputs.values())
        # (batch, 3, 320, 320)
        x1 = x + 1
        x2 = x - 1
        y = y + 1

        if self.training:
            print('训练模式')
            x1 = x2 * x1
        else:
            print('验证模式')
            x1 = x2 + x1

        # 分支判断
        if torch.onnx.is_in_onnx_export():
            print('开始onnx')

        # 切断tracing
        # if torch.onnx.isinonnxexport():
        #     # cast countmat to constant while exporting to ONNX
        #     countmat = torch.fromnumpy(
        #         countmat.cpu().detach().numpy()).to(device=img.device)

        ''' --------- 不支持 --------- '''
        # x = x[x > 0.5] # 不支持
        # x[0] = 1 # 不支持
        # x = torch.index_fill(x, -1, torch.arange(3), 1)  # 不支持
        # 不支持adaptiveAvgPool2d, Expand, ReLU6

        ''' --------- 支持 --------- '''
        # 支持float32 及 int
        # val_det = torch.tensor([1, 2], dtype=torch.float).view(1, 1, 1, 2)
        # index = torch.zeros_like(val_det, dtype=torch.long)
        # x = x.scatter(0, index, val_det)

        # x = x.max(-1)[0]
        # x = x.topk(3, dim=-1)[0]

        # return x1
        x2 = x1[:, :2, :, :] * y[0, 1]  # 支持
        return x1, x2
        # return x1, x2


if __name__ == '__main__':
    torch.set_printoptions(linewidth=320, sci_mode=False, precision=5, profile='long')
    np.set_printoptions(linewidth=320, suppress=True, precision=5, formatter={'float_kind': '{:11.5g}'.format})
    seed = 0
    # 随机种子
    np.random.seed(seed)
    random.seed(seed)
    # from yolo5 Speed-reproducibility tradeoff https://pytorch.org/docs/stable/notes/randomness.html
    torch.manual_seed(seed)

    file_onnx = 't_model.onnx'
    size_wh_input = (320, 320)

    model = TModel()
    # model = models.resnet18(pretrained=False)
    # model.eval()
    model.train()
    input_names = ['input0', 'input1']
    output_names = ['output0', 'output1']

    batch = 1
    num_end = 5
    # data_ts = torch.rand((batch, 3, *size_wh_input), dtype=torch.float)
    data_ts = torch.rand((1, 3, 230, 230), dtype=torch.float)
    # data_ts = [data_ts, torch.tensor([1, 2, 3])]
    data_dict = {input_names[0]: data_ts,
                 input_names[1]: torch.tensor([[1, 2, 3]])
                 }
    # data_np = data_ts.numpy()
    # data_np = np.random.uniform(0., 1, (1, 3, *size_wh_input), ).astype(np.float32)

    reses = model(data_dict)
    for res in reses:
        print(res.reshape(batch, -1)[:, :num_end])

    # 单输入
    # dynamic_axes = None
    dynamic_axes = {
        input_names[0]: {0: 'batch_size', 1: 'channel', 2: "height", 3: 'width'},  # 支持部份动态
        input_names[1]: {0: 'batch_size', 1: 'channel', 2: "height", 3: 'width'},  # 支持部份动态
    }

    # 多输出
    # dynamic_axes = {
    #     'input': {0: 'batch_size', 1: 'channel', 2: "height", 3: 'width'},
    #     'output0': {0: 'batch_size', 1: 'feature_maps'},
    #     'output1': {0: 'batch_size', 1: 'feature_maps'},
    #     'output2': {0: 'batch_size', 1: 'feature_maps'}
    # }

    torch.onnx.export(model, data_dict, file_onnx, verbose=False,
                      input_names=input_names,
                      output_names=output_names,
                      dynamic_axes=dynamic_axes,
                      )

    ''' --------------- 验证 ----------- '''
    # model = onnx.load(file_onnx)
    # 基本无用 Check that the IR is well formed
    # onnx.checker.check_model(model)
    # Print a human readable representation of the graph
    # print(onnx.helper.printable_graph(model.graph))

    onnx_session = onnxruntime.InferenceSession(file_onnx)
    output_name1 = onnx_session.get_outputs()[0].name
    output_name2 = onnx_session.get_outputs()[1].name
    input_name1 = onnx_session.get_inputs()[0].name
    input_name2 = onnx_session.get_inputs()[1].name

    print('开始')
    # 输出是list
    res_onnx = onnx_session.run(
        output_names=[output_name1, output_name2],
        input_feed={input_name1: data_ts.numpy(), input_name2: np.array([[1, 2, 3]], dtype=np.int64)})

    for res in res_onnx:
        print(res.reshape(batch, -1)[:, :num_end])
    print('debug')
    pass



'''
独立的颜色处理 与target 无关
'''
import numpy as np
import cv2
import random


def random_brightness(img, delta):
    img += random.uniform(-delta, delta)
    return img


def random_contrast(img, alpha_low, alpha_up):
    img *= random.uniform(alpha_low, alpha_up)
    return img


def random_saturation(img, alpha_low, alpha_up):
    hsv_img = cv2.cvtColor(img.astype(np.float32) / 255, cv2.COLOR_BGR2HSV)
    hsv_img[..., 1] *= random.uniform(alpha_low, alpha_up)
    img = cv2.cvtColor(hsv_img, cv2.COLOR_HSV2BGR) * 255
    # cv2.imshow('img', img/255)
    return img


def normalize(meta, mean, std):
    img = meta['img'].astype(np.float32)
    mean = np.array(mean, dtype=np.float64).reshape(1, -1)
    stdinv = 1 / np.array(std, dtype=np.float64).reshape(1, -1)
    cv2.subtract(img, mean, img)
    cv2.multiply(img, stdinv, img)
    meta['img'] = img
    return meta


def _normalize(img, mean, std):
    mean = np.array(mean, dtype=np.float32).reshape(1, 1, 3) / 255
    std = np.array(std, dtype=np.float32).reshape(1, 1, 3) / 255
    img = (img - mean) / std
    return img


def color_aug_and_norm(meta):
    img = meta['img'].astype(np.float32) / 255

    if random.randint(0, 1):
        # 0.2 亮度
        img = random_brightness(img, 0.2)

    if random.randint(0, 1):
        # [0.8, 1.2] 对比度
        img = random_contrast(img, 0.8, 1.2)

    if random.randint(0, 1):
        # [0.8, 1.2] 饱和度
        img = random_saturation(img, 0.8, 1.2)
    # cv2.imshow('trans', img)
    # cv2.waitKey(0)
    img = _normalize(img, [103.53, 116.28, 123.675], [57.375, 57.12, 58.395])
    return img



import argparse
import os

import cv2
import numpy as np

import onnxruntime
import torch

from ftools.ai.datas.dta_heighten.f_data_heighten4target import ftransform_ssd_test
from ftools.ai.datas.z_dataloader import get_data_type3

if __name__ == '__main__':
    file_onnx = ''
    path_root_weight = r'M:\AI\weights\feadre\zz'
    _file_read_weight = os.path.join(path_root_weight, 'nanodet_coco_p34r29_1' + '.pth')
    device = torch.device('cpu')
    size_wh_input = (320, 320)

    class CFG:
        pass


    path_host = 'M:'
    mode = 'bbox'
    cfg = CFG()
    cfg.IS_TRAIN = False
    cfg.IS_VAL = False
    cfg.IS_TEST = True
    cfg.IS_MULTI_SCALE = False
    cfg.num_workers = 0
    data_transform = ftransform_ssd_test(size_wh_input)

    data_info, dataloader_train, dataloader_test = get_data_type3(
        cfg,
        path_host=path_host,
        mode_train=mode, batch_train=1
    )


    # session = onnxruntime.InferenceSession(file_onnx)
    # ort_inputs = {session.get_inputs()[0].name: img[None, :, :, :]}
    # output = session.run(None, ort_inputs)

    # predictions = demo_postprocess(output[0], input_shape, p6=args.with_p6)[0]

    # boxes = predictions[:, :4]
    # scores = predictions[:, 4:5] * predictions[:, 5:]
    #
    # boxes_xyxy = np.ones_like(boxes)
    # boxes_xyxy[:, 0] = boxes[:, 0] - boxes[:, 2] / 2.
    # boxes_xyxy[:, 1] = boxes[:, 1] - boxes[:, 3] / 2.
    # boxes_xyxy[:, 2] = boxes[:, 0] + boxes[:, 2] / 2.
    # boxes_xyxy[:, 3] = boxes[:, 1] + boxes[:, 3] / 2.
    # boxes_xyxy /= ratio
    # dets = multiclass_nms(boxes_xyxy, scores, nms_thr=0.65, score_thr=0.1)
    #
    # if dets is not None:
    #     final_boxes, final_scores, final_cls_inds = dets[:, :4], dets[:, 4], dets[:, 5]
    #     origin_img = vis(origin_img, final_boxes, final_scores, final_cls_inds,
    #                      conf=args.score_thr, class_names=COCO_CLASSES)
    #
    # mkdir(args.output_dir)
    # output_path = os.path.join(args.output_dir, args.image_path.split("/")[-1])
    # cv2.imwrite(output_path, origin_img)



import os
import random

import onnxruntime
import torch
from torchvision import models
import numpy as np

from ftools.ai.fits.fweight import load_weight
from ftools.f_general import fshow_time

torch.set_printoptions(linewidth=320, sci_mode=False, precision=5, profile='long')
np.set_printoptions(linewidth=320, suppress=True, precision=5, formatter={'float_kind': '{:11.5g}'.format})
seed = 0
np.random.seed(seed)
random.seed(seed)
torch.manual_seed(seed)


def show_reses(reses):
    for res in reses:
        print(res.reshape(batch, -1)[:, :num_end], '\n')
    print(str('-------------------------------------------------------------' + '\n'))


def f_script(trace_script, input_t):
    reses = trace_script(input_t)
    show_reses(reses)


def f_y(model, input_t):
    reses = model(input_t)
    show_reses(reses)


def f_onnx(onnx_session, input_dict, output_names_list):
    reses = onnx_session.run(
        output_names=output_names_list,
        input_feed=input_dict)
    show_reses(reses)


class CFG:
    pass


if __name__ == '__main__':
    batch = 5
    num_end = 5
    num_time = 1
    path_save_root = '../../../_temp/file'

    cfg = CFG()
    cfg.IS_TEST = True
    cfg.IS_VAL = False
    cfg.NUM_CLASSES = 80  # ***
    cfg.device = torch.device('cpu')
    cfg.save_weight_name = ''
    cfg.PRED_CONF_THR = 0.35  # 这些参数无数只用于创建
    cfg.PRED_NMS_THR = 0.6  # 这些参数无数只用于创建
    path_host = 'M:'
    cfg.PATH_SAVE_WEIGHT = os.path.join(path_host, r'/AI/weights/feadre')

    ''' --------------------- yolo1 ----------------------
        能够正常转换  
        script 报错 return self._forward_impl(x) '''
    # from yolov1.net.fcre_model import finit_model  # 这个成功
    # file_onnx = 'yolo1.onnx'
    # output_names = ['output0', 'output1', 'output2']

    ''' --------------------- nanodet ----------------------
        全不行 
    '''
    # from nanodet.net.fcre_model import finit_model

    # file_onnx = 'nanodet.onnx'
    # output_names = ['output0', 'output1']

    ''' --------------------- yolox ----------------------
        trace_trace可以转换  
        script: 报错 x = self.stems[k](x)
        onnx: RuntimeError: step!=1 is currently not supported  不支持二维切片
    '''
    from yolox.net.fcre_model import finit_model

    file_onnx = 'yolox.onnx'
    file_script = 'yolox.pt'
    _file_read_weight = os.path.join(cfg.PATH_SAVE_WEIGHT, 'zz/yolox_coco_nano' + '.pth')
    output_names = ['output0', 'output1', 'output2']

    ''' -------------------------------------------------------------- '''
    input_t = torch.rand((batch, 3, 320, 320))
    input_cre_model = torch.rand((1, 3, 320, 320))

    model_z = finit_model(cfg)
    load_weight(file_weight=_file_read_weight, model=model_z,
                optimizer=None, lr_scheduler=None,
                device=cfg.device, ffun=None)
    model_z.eval()
    model = model_z.net

    fshow_time(f_y, arg_list=[model, input_t], num_time=num_time)

    ''' ------------ script ------------ '''
    file_script = os.path.join(path_save_root, file_script)

    # trace_script = torch.jit.script(model, input_trace)
    trace_trace = torch.jit.trace(model, input_cre_model)

    trace_trace.save(file_script)
    trace_trace = torch.jit.load(file_script)
    fshow_time(f_script, arg_list=[trace_trace, input_t], num_time=num_time)

    ''' ------------ onnx ------------ '''
    input_names = ['input0', ]

    dynamic_axes = {
        input_names[0]: {0: 'batch_size', 1: 'channel', 2: "height", 3: 'width'},  # 支持部份动态
    }

    file_onnx = os.path.join(path_save_root, file_onnx)
    torch.onnx.export(model, input_cre_model,
                      f=file_onnx,
                      verbose=False,
                      input_names=input_names,
                      output_names=output_names,
                      dynamic_axes=dynamic_axes,
                      opset_version=11,
                      )
    onnx_session = onnxruntime.InferenceSession(file_onnx)
    output_name1 = onnx_session.get_outputs()[0].name
    input_name1 = onnx_session.get_inputs()[0].name

    # res_onnx = onnx_session.run(
    #     output_names=[output_name1],
    #     input_feed={input_name1: input_t.numpy()})
    # print(res_onnx)

    arg_list = [onnx_session, {input_name1: input_t.numpy()}, output_names]
    fshow_time(f_onnx, arg_list=arg_list, num_time=num_time)



import math
import sys
import time
from abc import abstractmethod

import numpy as np
import cv2
import torch
from torch import nn

from ftools.GLOBAL_LOG import flog
from ftools.ai.picture.f_show import f_show_od_np4plt_v3, draw_text_chinese4cv


def nms(boxes, scores, iou_threshold):
    ''' IOU大于0.5的抑制掉
         boxes (Tensor[N, 4])) – bounding boxes坐标. 格式：(ltrb
         scores (Tensor[N]) – bounding boxes得分
         iou_threshold (float) – IoU过滤阈值
     返回:NMS过滤后的bouding boxes索引（降序排列）
     '''
    return torch.ops.torchvision.nms(boxes, scores, iou_threshold)


def batched_nms(boxes_ltrb, scores, idxs, threshold_iou):
    '''
    多 labels nms
    :param boxes_ltrb: 拉平所有类别的box重复的 n*20类,4
    :param scores: torch.Size([16766])
    :param idxs:  真实类别index 通过手动创建匹配的 用于表示当前 nms的类别 用于统一偏移 技巧
    :param threshold_iou:float 0.5
    :return:
    '''
    if boxes_ltrb.numel() == 0:  # 维度全乘
        return torch.empty((0,), dtype=torch.int64, device=boxes_ltrb.DEVICE)

    # torchvision.ops.boxes.batched_nms(boxes, scores, lvl, nms_thresh)
    # 根据最大的一个值确定每一类的偏移
    max_coordinate = boxes_ltrb.max()  # 选出每个框的 坐标最大的一个值
    # idxs 的设备和 boxes 一致 , 真实类别index * (1+最大值) 则确保同类框向 左右平移 实现隔离
    offsets = idxs.to(boxes_ltrb) * (max_coordinate + 1)
    # boxes 加上对应层的偏移量后，保证不同类别之间boxes不会有重合的现象
    boxes_for_nms = boxes_ltrb + offsets[:, None]
    keep = nms(boxes_for_nms, scores, threshold_iou)
    return keep


def batch_nms_v1(ids_batch1, p_ltrb1, p_labels1, p_scores1, threshold_nms):
    '''
    这个是以 迭代每一个类, 每次出一个类的所有batch的结果,
    难以以每个批次选100进行加速 , 如果用nms的全部结果,则没有问题
    :param ids_batch1: [nn]
    :param p_ltrb1: [nn,4] float32
    :param p_labels1: [nn]
    :param p_scores1: [nn]
    :return:
    '''
    device = p_ltrb1.DEVICE
    p_labels_unique = p_labels1.unique()  # nn -> n
    ids_batch2, p_scores2, p_labels2, p_ltrb2 = [], [], [], []

    for lu in p_labels_unique:  # 迭代每个类别
        # 过滤类别
        _mask = p_labels1 == lu
        _ids_batch = ids_batch1[_mask]
        _p_scores = p_scores1[_mask]
        _p_ltrb = p_ltrb1[_mask]
        # 这里用batch去区分
        keep = batched_nms(_p_ltrb, _p_scores, _ids_batch, threshold_nms)
        _p_labels = p_labels1[_mask]

        # 每批100个
        ids_batch2.append(_ids_batch[keep])
        p_scores2.append(_p_scores[keep])
        p_labels2.append(_p_labels[keep])
        p_ltrb2.append(_p_ltrb[keep])

    ids_batch2 = torch.cat(ids_batch2, 0)
    p_scores2 = torch.cat(p_scores2, 0)
    p_labels2 = torch.cat(p_labels2, 0)
    p_ltrb2 = torch.cat(p_ltrb2, 0)

    # print('12')
    return ids_batch2, p_ltrb2, p_labels2, p_scores2


def batch_nms_v2(ids_batch1, p_ltrb1, p_labels1, p_scores1, threshold_nms, num_max=99999):
    '''
    以每个batch进行迭代 可以每个batch选100个最好的结果 进行评分加速 和筛选
    :param ids_batch1: [nn]
    :param p_ltrb1: [nn,4] float32
    :param p_labels1: [nn]
    :param p_scores1: [nn]
    :param num_max: 通常一张图选 100个高分加速
    :return:
    '''
    # device = p_ltrb1.device
    ids_batch_unique = ids_batch1.unique()  # nn -> n
    ids_batch2, p_scores2, p_labels2, p_ltrb2 = [], [], [], []

    for lu in ids_batch_unique:  # 迭代每个批次
        # 过滤类别
        _mask = ids_batch1 == lu
        _p_scores = p_scores1[_mask]
        _p_labels = p_labels1[_mask]
        _p_ltrb = p_ltrb1[_mask]
        keep = batched_nms(_p_ltrb, _p_scores, _p_labels, threshold_nms)
        _ids_batch = ids_batch1[_mask]
        # 每批100个
        keep = keep[:num_max]

        ids_batch2.append(_ids_batch[keep])
        p_scores2.append(_p_scores[keep])
        p_labels2.append(_p_labels[keep])
        p_ltrb2.append(_p_ltrb[keep])

    ids_batch2 = torch.cat(ids_batch2, 0)
    p_scores2 = torch.cat(p_scores2, 0)
    p_labels2 = torch.cat(p_labels2, 0)
    p_ltrb2 = torch.cat(p_ltrb2, 0)

    # print('12')
    return ids_batch2, p_ltrb2, p_labels2, p_scores2


class Predicting_Base(nn.Module):
    '''
    由 f_fit_eval_base 运行至net模型主Center 再由模型主Center调用
    返回到 f_fit_eval_base
    '''

    def __init__(self, cfg) -> None:
        '''

        cfg.PRED_CONF_THR  # 阀值
        cfg.PRED_NMS_THR  # 阀值
        cfg.PRED_SELECT_TOPK  # 初选个数
        '''
        super(Predicting_Base, self).__init__()
        self.cfg = cfg

    @abstractmethod
    def p_init(self, model_outs, targets):
        '''
        p_init 处理数据
        :param model_outs:
        :param targets: 这个需要处理  例如 to.显卡
        :return:
        '''

    @abstractmethod
    def get_pscores(self, outs, targets):
        '''

        :param outs: p_init 处理后
        :param targets: 这个需要处理  例如 to.显卡
        :return:
            pscores : 这个用于判断 生成 mask_pos
            plabels : 这个用于二阶段
            pconf : 用于显示conf的统计值  这个可以随便返
        '''

    @abstractmethod
    def get_stage_res(self, outs, mask_pos, pscores, plabels, imgs_ts_4d, targets, off_ltrb_ts):
        '''

        :param imgs_ts_4d: 这个需要处理
        :param targets:
        :param off_ltrb_ts: 实际归一化尺寸 这个需要处理
        :return:
            ids_batch1 : 返回第二阶段
            pboxes_ltrb1 :
            plabels1 :
            pscores1 :
        '''

    def forward(self, model_outs, imgs_ts_4d=None, targets=None, off_ltrb_ts=None):
        '''
        由 FModelBase 已处理  toones_wh_ts_input
        :param model_outs: 这个输出就已经带了 device
        :return:
        '''
        ''' 进行数据的处理 传递到 get_pscores get_stage_res 中'''
        outs = self.p_init(model_outs, targets)

        ''' 返回分数 和 plabels    pconf可返回任意值 用于没有目标时分析值域 '''
        pscores, plabels, pconf = self.get_pscores(outs, targets)
        # batch, dim = pscores.shape
        mask_pos = pscores > self.cfg.PRED_CONF_THR
        if not torch.any(mask_pos):  # 如果没有一个对象  这个没满足  其它就不用走了
            print('该批次没有找到目标 max:{0:.2f} min:{0:.2f} mean:{0:.2f}'.format(pconf.max().item(),
                                                                          pconf.min().item(),
                                                                          pconf.mean().item(),
                                                                          ))
            return [None] * 5
            # return [torch.tensor(math.nan)] * 5

        # dim中 取500个
        if pscores.shape[-1] > self.cfg.PRED_SELECT_TOPK:  # 并行取top100 与mask_pos 进行and操作
            # 最大1000个
            ids_topk = pscores.topk(self.cfg.PRED_SELECT_TOPK, dim=-1)[1]  # torch.Size([32, 1000])
            mask_topk = torch.zeros_like(mask_pos)
            mask_topk[torch.arange(ids_topk.shape[0])[:, None], ids_topk] = True
            mask_pos = torch.logical_and(mask_pos, mask_topk)

        # ids_batch1, pboxes_ltrb1, plabels1, pscores1,pkp_keypoint
        reses = self.get_stage_res(outs, mask_pos, pscores, plabels, imgs_ts_4d, targets, off_ltrb_ts)

        # if self.cfg.MODE_VIS == 'bbox':  # 单人脸
        assert len(reses) == 4, 'res = self.get_stage_res(outs, mask_pos, pscores, plabels) 数据错误'
        ids_batch2, p_boxes_ltrb2, p_labels2, p_scores2 = batch_nms_v2(reses[0],
                                                                       reses[1],
                                                                       reses[2],
                                                                       reses[3],
                                                                       self.cfg.PRED_NMS_THR,
                                                                       num_max=self.cfg.PRED_NMS_KEEP,
                                                                       )

        return ids_batch2, p_boxes_ltrb2, None, p_labels2, p_scores2,
        # elif self.cfg.MODE_VIS == 'kps':
            # assert len(reses) == 5, 'res = self.get_stage_res(outs, mask_pos, pscores, plabels) 数据错误'
            # ids_batch2, p_boxes_ltrb2, p_keypoints2, p_labels2, p_scores2 = batch_nms4kp(
            #     ids_batch1=reses[0],
            #     p_boxes_ltrb1=reses[1],
            #     p_labels1=reses[2],
            #     p_scores1=reses[3],
            #     p_keypoints1=reses[4],
            #     threshold_nms=self.cfg.PRED_NMS_THR)
            # return ids_batch2, p_boxes_ltrb2, p_keypoints2, p_labels2, p_scores2
            # raise Exception('不支持 keypoints')
        # else:
        #     raise Exception('self.mode_vis = %s 错误' % self.cfg.mode_vis)



import cv2
import numpy as np
import dlib
import time
import math

detector = dlib.get_frontal_face_detector()
predictor = dlib.shape_predictor(r"M:\AI\datas\face68\shape_predictor_68_face_landmarks.dat")
POINTS_NUM_LANDMARK = 68


# 获取最大的人脸
def _largest_face(dets):
    if len(dets) == 1:
        return 0

    face_areas = [(det.right() - det.left()) * (det.bottom() - det.top()) for det in dets]

    largest_area = face_areas[0]
    largest_index = 0
    for index in range(1, len(dets)):
        if face_areas[index] > largest_area:
            largest_index = index
            largest_area = face_areas[index]

    print("largest_face index is {} in {} faces".format(largest_index, len(dets)))

    return largest_index


# 从dlib的检测结果抽取姿态估计需要的点坐标
def get_image_points_from_landmark_shape(landmark_shape):
    if landmark_shape.num_parts != POINTS_NUM_LANDMARK:
        print("ERROR:landmark_shape.num_parts-{}".format(landmark_shape.num_parts))
        return -1, None

    # 2D image points. If you change the image, you need to change vector
    image_points = np.array([
        (landmark_shape.part(30).x, landmark_shape.part(30).y),  # Nose tip
        (landmark_shape.part(8).x, landmark_shape.part(8).y),  # Chin
        (landmark_shape.part(36).x, landmark_shape.part(36).y),  # Left eye left corner
        (landmark_shape.part(45).x, landmark_shape.part(45).y),  # Right eye right corne
        (landmark_shape.part(48).x, landmark_shape.part(48).y),  # Left Mouth corner
        (landmark_shape.part(54).x, landmark_shape.part(54).y)  # Right mouth corner
    ], dtype="double")

    return 0, image_points


# 用dlib检测关键点，返回姿态估计需要的几个点坐标
def get_image_points(img):
    # gray = cv2.cvtColor( img, cv2.COLOR_BGR2GRAY )  # 图片调整为灰色
    dets = detector(img, 0)

    if 0 == len(dets):
        print("ERROR: found no face")
        return -1, None
    largest_index = _largest_face(dets)
    face_rectangle = dets[largest_index]

    landmark_shape = predictor(img, face_rectangle)

    return get_image_points_from_landmark_shape(landmark_shape)


# 获取翻转向量和平移向量
def get_pose_estimation(img_size, image_points):
    # 3D model points.
    model_points = np.array([
        (0.0, 0.0, 0.0),  # Nose tip 鼻子
        (0.0, -330.0, -65.0),  # Chin 下巴
        (-225.0, 170.0, -135.0),  # Left eye left corner
        (225.0, 170.0, -135.0),  # Right eye right corne 眼
        (-150.0, -150.0, -125.0),  # Left Mouth corner 嘴
        (150.0, -150.0, -125.0)  # Right mouth corner
    ])

    # Camera internals

    focal_length = img_size[1]  # hwc
    center = (img_size[1] / 2, img_size[0] / 2)
    camera_matrix = np.array(
        [[focal_length, 0, center[0]],
         [0, focal_length, center[1]],
         [0, 0, 1]], dtype="double"
    )

    print("Camera Matrix :{}".format(camera_matrix))

    dist_coeffs = np.zeros((4, 1))  # Assuming no lens distortion
    (success, rotation_vector, translation_vector) = cv2.solvePnP(
        model_points,  # (num_keypoints,3) 世界坐标 3D模型坐标
        image_points,  # (num_keypoints,2) 图像坐标系中点的坐标 xy
        camera_matrix,  # (3,3) 相机内参矩阵  通过标定
        dist_coeffs,  # (4,1) 畸变系数  通过标定
        flags=cv2.SOLVEPNP_ITERATIVE)

    print("Rotation Vector:\n {}".format(rotation_vector))
    print("Translation Vector:\n {}".format(translation_vector))
    return success, rotation_vector, translation_vector, camera_matrix, dist_coeffs


# 从翻转向量转换为欧拉角
def get_euler_angle(rotation_vector):
    # calculate rotation angles
    theta = cv2.norm(rotation_vector, cv2.NORM_L2)

    # transformed to quaterniond
    w = math.cos(theta / 2)
    x = math.sin(theta / 2) * rotation_vector[0][0] / theta
    y = math.sin(theta / 2) * rotation_vector[1][0] / theta
    z = math.sin(theta / 2) * rotation_vector[2][0] / theta

    ysqr = y * y
    # pitch (x-axis rotation)
    t0 = 2.0 * (w * x + y * z)
    t1 = 1.0 - 2.0 * (x * x + ysqr)
    print('t0:{}, t1:{}'.format(t0, t1))
    pitch = math.atan2(t0, t1)

    # yaw (y-axis rotation)
    t2 = 2.0 * (w * y - z * x)
    if t2 > 1.0:
        t2 = 1.0
    if t2 < -1.0:
        t2 = -1.0
    yaw = math.asin(t2)

    # roll (z-axis rotation)
    t3 = 2.0 * (w * z + x * y)
    t4 = 1.0 - 2.0 * (ysqr + z * z)
    roll = math.atan2(t3, t4)

    print('pitch:{}, yaw:{}, roll:{}'.format(pitch, yaw, roll))

    # 单位转换：将弧度转换为度
    Y = int((pitch / math.pi) * 180)
    X = int((yaw / math.pi) * 180)
    Z = int((roll / math.pi) * 180)

    return 0, Y, X, Z


def get_pose_estimation_in_euler_angle(landmark_shape, im_szie):
    try:
        ret, image_points = get_image_points_from_landmark_shape(landmark_shape)
        if ret != 0:
            print('get_image_points failed')
            return -1, None, None, None

        ret, rotation_vector, translation_vector, camera_matrix, dist_coeffs = get_pose_estimation(im_szie,
                                                                                                   image_points)
        if ret != True:
            print('get_pose_estimation failed')
            return -1, None, None, None

        ret, pitch, yaw, roll = get_euler_angle(rotation_vector)
        if ret != 0:
            print('get_euler_angle failed')
            return -1, None, None, None

        euler_angle_str = 'Y:{}, X:{}, Z:{}'.format(pitch, yaw, roll)
        print(euler_angle_str)
        return 0, pitch, yaw, roll

    except Exception as e:
        print('get_pose_estimation_in_euler_angle exception:{}'.format(e))
        return -1, None, None, None


if __name__ == '__main__':
    '''
    将旋转矩阵转化为4元数，再转化为欧拉角（弧度）
    '''
    # rtsp://admin:ts123456@10.20.21.240:554
    cap = cv2.VideoCapture(0)
    while cap.isOpened():
        start_time = time.time()

        # Read Image
        ret, im = cap.read()
        if ret != True:
            print('read frame failed')
            continue
        size = im.shape

        if size[0] > 700:
            h = size[0] / 3
            w = size[1] / 3
            im = cv2.resize(im, (int(w), int(h)), interpolation=cv2.INTER_CUBIC)
            size = im.shape

        ret, image_points = get_image_points(im)
        if ret != 0:
            print('get_image_points failed')
            continue

        ret, rotation_vector, translation_vector, camera_matrix, dist_coeffs = get_pose_estimation(size, image_points)
        if ret != True:
            print('get_pose_estimation failed')
            continue
        used_time = time.time() - start_time
        print("used_time:{} sec".format(round(used_time, 3)))

        ret, pitch, yaw, roll = get_euler_angle(rotation_vector)
        euler_angle_str = 'Y:{}, X:{}, Z:{}'.format(pitch, yaw, roll)
        print(euler_angle_str)

        # Project a 3D point (0, 0, 1000.0) onto the image plane.
        # We use this to draw a line sticking out of the nose

        (nose_end_point2D, jacobian) = cv2.projectPoints(np.array([(0.0, 0.0, 1000.0)]), rotation_vector,
                                                         translation_vector, camera_matrix, dist_coeffs)

        for p in image_points:
            cv2.circle(im, (int(p[0]), int(p[1])), 3, (0, 0, 255), -1)

        p1 = (int(image_points[0][0]), int(image_points[0][1]))
        p2 = (int(nose_end_point2D[0][0][0]), int(nose_end_point2D[0][0][1]))

        cv2.line(im, p1, p2, (255, 0, 0), 2)

        # Display image
        # cv2.putText( im, str(rotation_vector), (0, 100), cv2.FONT_HERSHEY_PLAIN, 1, (0, 0, 255), 1 )
        cv2.putText(im, euler_angle_str, (0, 120), cv2.FONT_HERSHEY_PLAIN, 1, (0, 0, 255), 1)
        cv2.imshow("Output", im)
        cv2.waitKey(1)
